/*****************************************************************************

This source code is made available to you under the terms of the Open
Development License.  This Package should be accompanied by a file that
contains the associated licensing terms.  If you are unable to find this file,
please refer to its location on the Internet at www.rocklyte.com/license.html.

******************************************************************************

###CLASS###
Name:      View
Version:   1.0
ID:        ID_VIEW
Status:    In Development
Category:  GUI
Date:      January 2004
Author:    Rocklyte Systems
Copyright: Rocklyte Systems, 2003-2004.  All rights reserved.
Keywords:  view, tree, File, Directory, Manager, Files, Directories, Management, Explorer
Short:     The View class is used to display XML data in a number of possible formats to the user.

###DESCRIPTION###
<p>The View is an interface class that is designed to ease the display and
management of multiple items within a container.  It is capable of displaying
items in a number of formats, including a simple list, tree view and column view.
The ability to sort data and handle different types of data such as date
and time information is provided.  The View class is commonly used for file
display and certain classes such as the <a href="fileview.html">FileView</a> are dependent on the
functionality that it provides.</p>

<p>Items are created and stored within each View as a large XML statement and
are managed via the XML class.  This simplifies the creation and retrieval
of item data and also allows you to develop hierarchies and tree structures
for complex item arrangement.</p>

<p>The View class inherits the fields and functionality of the <a href="xml.html">XML</a> class, which
you can use for scanning view items after they have been added.  The <a href="font.html">Font</a> class
is also inherited, allowing you to set details such as the default font face
through the view.</p>

<p>When the user double-clicks on a view item, the object will activate
itself. Once activated, the view object will send Activate actions on to any
children that have been initialised to the view.  This allows you to develop
a response for user interaction with the view.  If the sensitive option has
been enabled as one of the Flags attributes, activation will occur whenever
the user selects a file (thus single clicks will be enough to cause object
activation).</p>

###END###

*****************************************************************************/

//#define DEBUG

#include <pandora/system/all.h>
#include <pandora/graphics/all.h>
#include <pandora/files/all.h>
#include <pandora/misc/all.h>
#include <pandora/modules/screen.h>
#include <pandora/modules/strings.h>
#include <pandora/modules/render.h>
#include <pandora/input/joyport.h>
#include <pandora/input/keyboard.h>
#include <pandora/data/xml.h>
#include <pandora/data/config.h>

static ERROR CMDInit(OBJECTPTR, struct KernelBase *);
static ERROR CMDExpunge(void);

#define VER_VIEW  1.0

MODULE_HEADER = {
   MODULE_HEADER_V1, CMDInit, NULL, NULL, CMDExpunge,
   JMP_DEFAULT, 0, CPU_PC, VER_VIEW, VER_KERNEL,
   "Rocklyte Systems",
   "Rocklyte Systems (c) 2003-2004. All rights reserved.",
   "January 2004",
   "View"
};

static struct KernelBase  *KernelBase;
static struct ScreenBase *ScreenBase;
static struct StringsBase *StringsBase;
static struct RenderBase *RenderBase;
static OBJECTPTR ViewClass = 0;
static OBJECTPTR modStrings = NULL, modScreen = NULL, modRender = NULL;
static FIELD FID_PrivateDataSize;
static UBYTE glDateFormat[28] = "dd-mm-yy hh:nn";

#define MIN_COLWIDTH 6

/*****************************************************************************
** Object structure.
*/

struct View {
   OBJECT_HEADER         /* Class header */
   struct XML *XML;      /* The XML object contains all view information */
   struct Font *Font;    /* Font object */
   OBJECTID VScroll;     /* Reference to a vertical scroll bar */
   OBJECTID HScroll;     /* References to a horizontal scroll bar */
   OBJECTID DrawableID;  /* Drawable destination */
   OBJECTID InputID;     /* The keyboard to read information from */
   OBJECTID FocusID;     /* The drawable to monitor for the focus */
   LONG   Flags;         /* Special flags */
   LONG   Style;         /* Style of view */
   LONG   LeftMargin;    /* Left margin */
   LONG   TopMargin;     /* Top margin */
   LONG   RightMargin;   /* Right margin */
   LONG   BottomMargin;  /* Bottom margin */
   LONG   HSpacing;      /* Horizontal spacing between each item when arranged across the page */
   LONG   VSpacing;      /* Vertical spacing between each item when arranged down the page */
   LONG   SelectedTag;   /* The most recently selected tag [monitorable] */
   LONG   ActiveTag;     /* The most recently clicked or activated tag (not necessarily selected) */
   LONG   HighlightTag;  /* Current user-highlighted item */
   LONG   MaxItemWidth;  /* Maximum item width, in pixels (applies to certain view styles) */
   STRING ColumnString;  /* String defining the column headers */
   LONG   ColThickness;  /* Thickness of column button borders when in COLUMN mode */

   PRIVATE_FIELDS
   LONG   XPosition;         /* Horizontal scrolling offset */
   LONG   YPosition;         /* Vertical scrolling offset */
   LONG   Dimensions;        /* Dimension settings */
   FLOAT  Width, Height;     /* Dimension settings */
   FLOAT  XCoord, YCoord;    /* Dimension settings */
   FLOAT  XOffset, YOffset;  /* Dimension settings */
   LONG   Buttons;           /* Current button presses */
   LONG   ClickX, ClickY;
   LONG   ClickIndex;
   LONG   PageWidth;          /* Width of the scrollable page area */
   LONG   PageHeight;         /* Height of the scrollable page area */
   LONG   AreaX, AreaY, AreaWidth, AreaHeight; /* Dimensions of our area within the parent drawable */
   LONG   *SelectedTags;
   LONG   SelectionIndex;     /* Selection index, used to assist initialisation process only */
   struct RGB HighlightRGB;     /* Highlight colour on mouse over. */
   struct RGB SelectRGB;        /* Colour to use for selected items */
   struct RGB ItemRGB;          /* Default colour to apply to new items */
   struct RGB ColBackground;
   struct RGB ColHighlight;
   struct RGB ColShadow;
   struct RGB HairlineRGB;
   struct Column *Columns;      /* Column information */
   BYTE   DateFormat[28];       /* Preferred date format */
   UBYTE  Selection[120];     /* Currently selected file/directory */
   BYTE   Sort[4];              /* Sort indexes to each column (4 max) from most to least important.  A negative index means that the sort is descending */
   struct Column *ColumnResize; /* Set if a column is being resized */
   UBYTE  Deselect;
   BYTE   KeyboardChannel;    /* Set to TRUE if keyboard channel is active */
   BYTE   PointerLocked;        /* Set if the pointer is locked */
   WORD   LineHeight;
   WORD   IconWidth;          /* Maximum pixel width/height of icons */
   WORD   ColumnHeight;       /* Height of column buttons when in column mode */
   STRING Selections;
   struct {
      struct Picture *Picture;
      UBYTE String[40];
   } Cache[4];
};

/*****************************************************************************
** Node management
*/

struct ViewNode {
   struct Picture *Icon;
   struct RGB FontRGB;
   union { LONG X; LONG XCoord; };
   union { LONG Y; LONG YCoord; };
   WORD Width;
   WORD Height;
   LONG Flags;
};

#define NODE_SELECTED    0x00000001  /* Item is selected */
#define NODE_HIGHLIGHTED 0x00000002  /* Item is temporarily highlighted (e.g. the mouse pointer is lying over it) */
#define NODE_NEWCOLUMN   0x00000004  /* The node forms the top of a new column */
#define NODE_CACHEICON   0x00000008  /* The icon is from the cache */

/*****************************************************************************
** Column management
*/

struct Column {
   struct Column *Next;
   UBYTE Name[20];
   UBYTE Text[32];
   BYTE  Type;         /* The type of data displayed in the column */
   BYTE  Sort;         /* Sort order */
   WORD  Flags;        /* Column flags */
   WORD  Width;        /* Pixel width */
   struct RGB Colour;  /* Background colour */
};

#define CF_COLOUR     0x0001
#define CF_SHOWICONS  0x0002 /* Show icons in this column (NB: will be overridden if VWF_NOICONS is in use) */
#define CF_RIGHTALIGN 0x0004
#define CF_DELETE     0x0008 /* Marked for deletion (internal use only) */

enum {
   SORT_UNSORTED = 0,
   SORT_ASCENDING,
   SORT_DESCENDING
};

enum {
   CT_VARIANT=1,
   CT_BYTESIZE,
   CT_NUMERIC,
   CT_DATE
};

/*****************************************************************************
** Class definition.
*/

static struct FieldDef ViewFlags[] = {
   { "MULTISELECT",  VWF_MULTISELECT },
   { "NOICONS",      VWF_NOICONS },
   { "SENSITIVE",    VWF_SENSITIVE },
   { "PRESERVEBKGD", VWF_PRESERVEBKGD },
   { "INVISIBLE",    VWF_INVISIBLE },
   { "HAIRLINES",    VWF_HAIRLINES },
   { "DRAWTABLE",    VWF_DRAWTABLE },
   { "NOSELECT",     VWF_NOSELECT },
   { NULL, NULL }
};

static struct FieldDef ViewStyle[] = {
   { "TREE",     VIEW_TREE },
   { "LIST",     VIEW_LIST },
   { "LONGLIST", VIEW_LONGLIST },
   { "ICON",     VIEW_ICON },
   { "COLUMN",   VIEW_COLUMN },
   { NULL, NULL }
};

static ERROR GET_DateFormat(struct View *, STRING *);
static ERROR GET_Height(struct View *, struct Variable *);
static ERROR GET_Selection(struct View *, STRING *);
static ERROR GET_SelectionIndex(struct View *, LONG *);
static ERROR GET_SelectedTags(struct View *, LONG **);
static ERROR GET_TotalSelected(struct View *, LONG *);
static ERROR GET_Width(struct View *, struct Variable *);
static ERROR GET_X(struct View *, struct Variable *);
static ERROR GET_XOffset(struct View *, struct Variable *);
static ERROR GET_Y(struct View *, struct Variable *);
static ERROR GET_YOffset(struct View *, struct Variable *);

static ERROR SET_BorderOffset(struct View *, LONG *);
static ERROR SET_DateFormat(struct View *, STRING);
static ERROR SET_ColBackground(struct View *, STRING);
static ERROR SET_ColHighlight(struct View *, STRING);
static ERROR SET_ColShadow(struct View *, STRING);
static ERROR SET_Columns(struct View *, STRING);
static ERROR SET_Hairline(struct View *, STRING);
static ERROR SET_Height(struct View *, struct Variable *);
static ERROR SET_HighlightColour(struct View *, STRING);
static ERROR SET_HScroll(struct View *, OBJECTID *);
static ERROR SET_ItemColour(struct View *, STRING);
static ERROR SET_SelectColour(struct View *, STRING);
static ERROR SET_Selection(struct View *, STRING);
static ERROR SET_SelectionIndex(struct View *, LONG *);
static ERROR SET_Style(struct View *, LONG *);
static ERROR SET_VScroll(struct View *, OBJECTID *);
static ERROR SET_Width(struct View *, struct Variable *);
static ERROR SET_X(struct View *, struct Variable *);
static ERROR SET_XOffset(struct View *, struct Variable *);
static ERROR SET_Y(struct View *, struct Variable *);
static ERROR SET_YOffset(struct View *, struct Variable *);

static struct FieldArray ViewFields[] = {
   { "XML",          0, FDF_CHILD|FDF_R,            ID_XML,      NULL, NULL },
   { "Font",         0, FDF_CHILD|FDF_R,            ID_FONT,     NULL, NULL },
   { "VScroll",      0, FDF_OBJECTID|FDF_RW,        ID_SCROLL,   NULL, SET_VScroll },
   { "HScroll",      0, FDF_OBJECTID|FDF_RW,        ID_SCROLL,   NULL, SET_HScroll },
   { "Drawable",     0, FDF_OBJECTID|FDF_RW,        ID_RENDER,   NULL, NULL },
   { "Input",        0, FDF_OBJECTID|FDF_RI,        ID_KEYBOARD, NULL, NULL },
   { "Focus",        0, FDF_OBJECTID|FDF_RI,        0,           NULL, NULL },
   { "Flags",        0, FDF_LONGFLAGS|FDF_RW,       (LONG)&ViewFlags, NULL, NULL },
   { "Style",        0, FDF_LONG|FDF_LOOKUP|FDF_RW, (LONG)&ViewStyle, NULL, SET_Style },
   { "LeftMargin",   0, FDF_LONG|FDF_RI,            0,         NULL, NULL },
   { "TopMargin",    0, FDF_LONG|FDF_RI,            0,         NULL, NULL },
   { "RightMargin",  0, FDF_LONG|FDF_RI,            0,         NULL, NULL },
   { "BottomMargin", 0, FDF_LONG|FDF_RI,            0,         NULL, NULL },
   { "HSpacing",     0, FDF_LONG|FDF_RW,            0,         NULL, NULL },
   { "VSpacing",     0, FDF_LONG|FDF_RW,            0,         NULL, NULL },
   { "SelectedTag",  0, FDF_LONG|FDF_RW,            0,         NULL, NULL },
   { "ActiveTag",    0, FDF_LONG|FDF_R,             0,         NULL, NULL },
   { "HighlightTag", 0, FDF_LONG|FDF_R,             0,         NULL, NULL },
   { "MaxItemWidth", 0, FDF_LONG|FDF_RW,            0,         NULL, NULL },
   { "Columns",      0, FDF_STRING|FDF_RW,          0,         NULL, SET_Columns },
   { "ColThickness", 0, FDF_LONG|FDF_RW,            0,         NULL, NULL },

   /** Virtual fields ***/
   { "BorderOffset",   0, FDF_LONG|FDF_W,            0, NULL, SET_BorderOffset },
   { "DateFormat",     0, FDF_STRING|FDF_RW,         0, GET_DateFormat, SET_DateFormat },
   { "Selection",      0, FDF_STRING|FDF_RW,         0, GET_Selection, SET_Selection },
   { "SelectionIndex", 0, FDF_LONG|FDF_RW,           0, GET_SelectionIndex, SET_SelectionIndex },
   { "SelectedTags",   0, FDF_POINTER|FDF_R,         0, GET_SelectedTags, NULL },
   { "TotalSelected",  0, FDF_LONG|FDF_R,            0, GET_TotalSelected, NULL },
   { "View",           0, FDF_LONG|FDF_LOOKUP|FDF_W, (LONG)&ViewStyle, NULL, SET_Style },
   { "HighlightColour",0, FDF_STRING|FDF_W,          0, NULL, SET_HighlightColour },
   { "SelectColour",   0, FDF_STRING|FDF_W,          0, NULL, SET_SelectColour },
   { "ItemColour",     0, FDF_STRING|FDF_W,          0, NULL, SET_ItemColour },
   { "ColBackground",  0, FDF_STRING|FDF_W,          0, NULL, SET_ColBackground },
   { "ColHighlight",   0, FDF_STRING|FDF_W,          0, NULL, SET_ColHighlight },
   { "ColShadow",      0, FDF_STRING|FDF_W,          0, NULL, SET_ColShadow },
   { "Hairline",       0, FDF_STRING|FDF_W,          0, NULL, SET_Hairline },

   /*** Variable fields ***/
   { "Height",  0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_Height,  SET_Height },
   { "Width",   0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_Width,   SET_Width },
   { "X",       0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_X,       SET_X },
   { "XCoord",  0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_X,       SET_X },
   { "XOffset", 0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_XOffset, SET_XOffset },
   { "Y",       0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_Y,       SET_Y },
   { "YCoord",  0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_Y,       SET_Y },
   { "YOffset", 0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_YOffset, SET_YOffset },
   END_FIELD
};

/*****************************************************************************
** Actions
*/

static ERROR VIEW_ActionNotify(struct View *, struct acActionNotify *);
static ERROR VIEW_Activate(struct View *, APTR);
static ERROR VIEW_Clear(struct View *, APTR);
static ERROR VIEW_ClosingTag(struct View *, APTR);
static ERROR VIEW_DataChannel(struct View *, struct acDataChannel *);
static ERROR VIEW_Disable(struct View *, APTR);
static ERROR VIEW_Draw(struct View *, struct acDraw *);
static ERROR VIEW_Enable(struct View *, APTR);
static ERROR VIEW_Focus(struct View *, APTR);
static ERROR VIEW_Free(struct View *, APTR);
static ERROR VIEW_GetUnlistedField(struct View *, struct acGetUnlistedField *);
static ERROR VIEW_Hide(struct View *, APTR);
static ERROR VIEW_Init(struct View *, APTR);
static ERROR VIEW_LostFocus(struct View *, APTR);
static ERROR VIEW_NewObject(struct View *, APTR);
static ERROR VIEW_Refresh(struct View *, APTR);
static ERROR VIEW_ScrollToPoint(struct View *, struct acScrollToPoint *);
static ERROR VIEW_Show(struct View *, APTR);
static ERROR VIEW_Sort(struct View *, APTR);
static ERROR VIEW_UserClick(struct View *, struct acUserClick *);
static ERROR VIEW_UserClickRelease(struct View *, struct acUserClickRelease *);
static ERROR VIEW_UserMovement(struct View *, struct acUserMovement *);

static struct ActionArray ViewActions[] = {
   { AC_ActionNotify,     VIEW_ActionNotify },
   { AC_Activate,         VIEW_Activate },
   { AC_Clear,            VIEW_Clear },
   { AC_ClosingTag,       VIEW_ClosingTag },
   { AC_DataChannel,      VIEW_DataChannel },
   { AC_Disable,          VIEW_Disable },
   { AC_Draw,             VIEW_Draw },
   { AC_Enable,           VIEW_Enable },
   { AC_Focus,            VIEW_Focus },
   { AC_Free,             VIEW_Free },
   { AC_GetUnlistedField, VIEW_GetUnlistedField },
   { AC_Hide,             VIEW_Hide },
   { AC_Init,             VIEW_Init },
   { AC_LostFocus,        VIEW_LostFocus },
   { AC_NewObject,        VIEW_NewObject },
   { AC_Refresh,          VIEW_Refresh },
   { AC_ScrollToPoint,    VIEW_ScrollToPoint },
   { AC_Show,             VIEW_Show },
   { AC_Sort,             VIEW_Sort },
   { AC_UserClick,        VIEW_UserClick },
   { AC_UserClickRelease, VIEW_UserClickRelease },
   { AC_UserMovement,     VIEW_UserMovement },
   { NULL, NULL }
};

/*****************************************************************************
** Methods
*/

static ERROR VIEW_CacheIcon(struct View *, struct mtCacheViewIcon *);
static ERROR VIEW_InvertSelection(struct View *, APTR);
static ERROR VIEW_RemoveItem(struct View *, struct mtRemoveViewItem *);
static ERROR VIEW_RemoveTag(struct View *, struct mtRemoveViewTag *);
static ERROR VIEW_SelectAll(struct View *, APTR);
static ERROR VIEW_SelectItem(struct View *, struct mtSelectItem *);
static ERROR VIEW_SelectNone(struct View *, APTR);
static ERROR VIEW_SetItem(struct View *, struct mtSetViewItem *);
static ERROR VIEW_SortColumn(struct View *, struct mtSortColumn *);
static ERROR VIEW_SortColumnIndex(struct View *, struct mtSortColumnIndex *);

struct FunctionField argsCacheIcon[]       = { { "Location", ARG_STR }, { NULL, NULL } };
struct FunctionField argsRemoveItem[]      = { { "ID",  ARG_LONG }, { "TagIndex", ARG_LONG }, { NULL, NULL } };
struct FunctionField argsRemoveTag[]       = { { "TagIndex", ARG_LONG }, { NULL, NULL } };
struct FunctionField argsSelectItem[]      = { { "Tag", ARG_STR }, { "Item", ARG_STR }, { NULL, NULL } };
struct FunctionField argsSetItem[]         = { { "ID",  ARG_LONG }, { "TagIndex", ARG_LONG }, { "Tag", ARG_STR }, { "Attrib", ARG_STR }, { "Value", ARG_STR }, { NULL, NULL } };
struct FunctionField argsSortColumn[]      = { { "Column", ARG_STR }, { "Descending", ARG_LONG }, { NULL, NULL } };
struct FunctionField argsSortColumnIndex[] = { { "Column", ARG_LONG }, { "Descending", ARG_LONG }, { NULL, NULL } };

static struct MethodArray ViewMethods[] = {
   { MT_CacheViewIcon,   VIEW_CacheIcon,       "CacheIcon",       argsCacheIcon, sizeof(struct mtCacheViewIcon) },
   { MT_InvertSelection, VIEW_InvertSelection, "InvertSelection", NULL, NULL },
   { MT_RemoveViewItem,  VIEW_RemoveItem,      "RemoveItem",      argsRemoveItem, sizeof(struct mtRemoveViewItem) },
   { MT_RemoveViewTag,   VIEW_RemoveTag,       "RemoveTag",       argsRemoveTag,  sizeof(struct mtRemoveViewTag) },
   { MT_SelectAll,       VIEW_SelectAll,       "SelectAll",       NULL, NULL },
   { MT_SelectItem,      VIEW_SelectItem,      "SelectItem",      argsSelectItem,      sizeof(struct mtSelectItem) },
   { MT_SortColumn,      VIEW_SortColumn,      "SortColumn",      argsSortColumn,      sizeof(struct mtSortColumn) },
   { MT_SortColumnIndex, VIEW_SortColumnIndex, "SortColumnIndex", argsSortColumnIndex, sizeof(struct mtSortColumnIndex) },
   { MT_SetViewItem,     VIEW_SetItem,         "SetItem",         argsSetItem,         sizeof(struct mtSetViewItem) },
   { MT_SelectNone,      VIEW_SelectNone,      "SelectNone",      NULL, NULL },
   { NULL, NULL, NULL, NULL }
};

/*****************************************************************************
** Prototypes
*/

static void ArrangeItems(struct View *);
static void CheckSelectedItems(struct View *, struct XMLTag *);
static void CheckPointerCursor(struct View *, WORD, WORD);
static void CheckItemVisible(struct View *, struct XMLTag *);
static ERROR CalculateHScroll(struct View *);
static ERROR CalculateVScroll(struct View *);
static void DrawItem(struct View *, struct XMLTag *);
static void DrawView(struct View *, struct Render *, struct Bitmap *);
static STRING GetXMLAttribute(struct XMLTag *, STRING);
static ERROR SortItems(struct View *);
static BYTE SelectItem(struct View *, struct XMLTag *, BYTE);
static ERROR SetIcon(struct View *, struct XMLTag *, STRING);

/*****************************************************************************
** Command: Init()
*/

static ERROR CMDInit(OBJECTPTR argModule, struct KernelBase *argKernelBase)
{
   struct mtReadConfig read;
   OBJECTPTR config;

   KernelBase = argKernelBase;

   if (ResolveFields("PrivateDataSize", &FID_PrivateDataSize,
                     TAGEND) != ERR_Okay) {
      return(ObjectError(argModule, ERH_InitModule, ERR_ResolveFields));
   }

   /*** Open required modules ***/

   if (CreateObject(ID_MODULE, NULL, &modScreen, NULL,
      FID_Name|TSTRING,   "screen",
      FID_Version|TFLOAT, MODVERSION_SCREEN,
      TAGEND) IS ERR_Okay) {
      if (GetField(modScreen, FID_ModBase, FT_POINTER, &ScreenBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));

   if (CreateObject(ID_MODULE, NULL, &modStrings, NULL,
      FID_Name|TSTRING,   "strings",
      FID_Version|TFLOAT, MODVERSION_STRINGS,
      TAGEND) IS ERR_Okay) {
      if (GetField(modStrings, FID_ModBase, FT_POINTER, &StringsBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));

   if (CreateObject(ID_MODULE, NULL, &modRender, NULL,
      FID_Name|TSTRING,   "render",
      FID_Version|TFLOAT, MODVERSION_RENDER,
      TAGEND) IS ERR_Okay) {
      if (GetField(modRender, FID_ModBase, FT_POINTER, &RenderBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));

   if (CreateObject(ID_CONFIG, NULL, &config, NULL,
      FID_Location|TSTRING, "user:config/locale.cfg",
      TAGEND) IS ERR_Okay) {
      read.Section = "Locale";
      read.Item = "FileDate";
      if (Action(MT_ReadConfig, config, &read) IS ERR_Okay) {
         if (read.Data[0]) StrCopy(read.Data, glDateFormat, sizeof(glDateFormat));
      }
      acFree(config);
   }

   return(CreateObject(ID_CLASS, NULL, (OBJECTPTR *)&ViewClass, NULL,
      FID_BaseClassID|TLONG, ID_VIEW,
      FID_Version|TFLOAT,    VER_VIEW,
      FID_Name|TSTRING,      "View",
      FID_Category|TLONG,    CCF_GUI,
      FID_Flags|TLONG,       CLF_PROMOTECHILDREN|CLF_PRIVATEONLY,
      FID_Actions|TPTR,      ViewActions,
      FID_Methods|TPTR,      ViewMethods,
      FID_Fields|TPTR,       ViewFields,
      FID_Size|TLONG,        sizeof(struct View),
      TAGEND));
}

/*****************************************************************************
** Command: Expunge()
*/

static ERROR CMDExpunge(void)
{
   if (ViewClass)  { acFree(ViewClass);  ViewClass = NULL; }
   if (modScreen)  { acFree(modScreen);  modScreen = NULL; }
   if (modStrings) { acFree(modStrings); modStrings = NULL; }
   if (modRender)  { acFree(modRender);  modRender = NULL; }
   return(ERR_Okay);
}

/*****************************************************************************
** Internal: InitKeyboard()
*/

static void InitKeyboard(struct View *Self)
{
   OBJECTPTR input;

   if (Self->KeyboardChannel IS FALSE) {
      if (AccessObject(Self->InputID, 5000, &input) IS ERR_Okay) {
         SubscribeChannel(input, Self);
         Self->KeyboardChannel = TRUE;
         ReleaseObject(input);
      }
      else ObjectError(Self, ERH_Focus, ERR_AccessObject);
   }
}

/*****************************************************************************
** View: ActionNotify
*/

static ERROR VIEW_ActionNotify(struct View *Self, struct acActionNotify *Args)
{
   if (Args->Error != ERR_Okay) return(ERR_Okay);

   if (Args->ActionID IS AC_Redimension) {
      /*** Recompute our dimensions ***/

      struct Variable var;

      var.Type = FD_LARGE;
      GET_X(Self, &var); Self->AreaX = var.Large;
      GET_Y(Self, &var); Self->AreaY = var.Large;
      GET_Width(Self, &var);  Self->AreaWidth  = var.Large;
      GET_Height(Self, &var); Self->AreaHeight = var.Large;

      /*** Rearrange/recalculate dimensions for all view items ***/

      ArrangeItems(Self);
   }
   else if (Args->ActionID IS AC_Disable) {
      DelayMsg(AC_Draw, Self->DrawableID, NULL);
   }
   else if (Args->ActionID IS AC_Enable) {
      DelayMsg(AC_Draw, Self->DrawableID, NULL);
   }
   else if (Args->ActionID IS AC_Focus) {
      InitKeyboard(Self);
   }
   else if (Args->ActionID IS AC_LostFocus) {
      Action(AC_LostFocus, Self, NULL);
   }
   else if (Args->ActionID IS AC_UserClick) {
      Action(AC_UserClick, Self, Args->Args);
   }
   else if (Args->ActionID IS AC_UserClickRelease) {
      Action(AC_UserClickRelease, Self, Args->Args);
   }
   else if (Args->ActionID IS AC_UserMovement) {
      Action(AC_UserMovement, Self, Args->Args);
   }

   return(ERR_Okay);
}

/*****************************************************************************
** Action: Activate
*/

static ERROR VIEW_Activate(struct View *Self, APTR Void)
{
   struct ChildEntry list[20];
   LONG count, i;

   //if ((Self->SelectedTag IS -1) AND (Self->ActiveTag IS -1)) return(ERR_Okay);

   DPrintF("~Activate()","[View:%d]", Self->Head.UniqueID);

   count = ARRAYSIZE(list);
   if (ListChildren(Self->Head.UniqueID, list, &count) IS ERR_Okay) {

      for (i=0; i < count; i++) ActionMsg(AC_Activate, list[i].ObjectID, NULL);
   }
   else {
      StepBack();
      return(ERR_ListChildren);
   }

   StepBack();
   return(ERR_Okay);
}

/******************************************************************************

###METHOD###
Name:  CacheIcon
Short: Caches icons in the View object.

###DESCRIPTION###
<p>The CacheIcon method is used for pre-caching commonly used icons in a
view object.  It is particularly useful in large views that repeat the same
icon for multiple items.  Pre-caching an icon can be useful if you clear the
view on a regular basis and do not want to re-load commonly used icons on
each refresh.</p>

<p>The icon must be referenced using the string format "category/icon".</p>

###INPUT###
STRING/Icon: The icon that you want to cache.  Use the format "category/icon".

###RESULT###
ERR_Okay: Operation successful.
ERR_Args: The Icon argument was not specified.
ERR_ArrayFull: The cache is at maximum capacity.

###END###

******************************************************************************/

static ERROR VIEW_CacheIcon(struct View *Self, struct mtCacheViewIcon *Args)
{
   struct RenderInfo info;
   struct Picture *picture;
   UBYTE buffer[80];
   WORD i;

   if ((!Args) OR (!Args->Icon)) return(ObjectError(Self, MT_CacheViewIcon, ERR_Args));

   if (StrCompare("icons:", Args->Icon, 6, NULL) IS ERR_Okay) {
      StrFormat(buffer, sizeof(buffer), "%s(16)", Args->Icon);
   }
   else StrFormat(buffer, sizeof(buffer), "icons:%s(16)", Args->Icon);

   DPrintF("~CacheIcon()","[View:%d] Icon: %s", Self->Head.UniqueID, buffer);

   /*** Check if the icon is already cached ***/

   for (i=0; i < ARRAYSIZE(Self->Cache); i++) {
      if (StrCompare(buffer, Self->Cache[i].String, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         DPrintF("CacheIcon:","Icon is already cached.");
         StepBack();
         return(ERR_Okay);
      }
   }

   /*** Find an empty cache entry ***/

   for (i=0; (i < ARRAYSIZE(Self->Cache)) AND (Self->Cache[i].Picture); i++);

   if (i >= ARRAYSIZE(Self->Cache)) {
      return(StepObjectError(Self, MT_CacheViewIcon, ERR_ArrayFull));
   }

   if (GetRenderInfo(Self->DrawableID, &info, sizeof(info)) != ERR_Okay) info.BitsPerPixel = 32;

   if (NewPrivateObject(ID_PICTURE, NF_CHILD, &picture) IS ERR_Okay) {

      SetFields(picture, FID_Location|TSTRING, buffer, TAGEND);
      picture->Bitmap->BitsPerPixel = info.BitsPerPixel;
      picture->Flags |= PCF_32BITALPHA|PCF_CACHE;
      if (acInit(picture) IS ERR_Okay) {
         Self->Cache[i].Picture = picture;
         StrCopy(buffer, Self->Cache[i].String, sizeof(Self->Cache[i].String));
      }
      else acFree(picture);
   }
   else return(StepObjectError(Self,MT_CacheViewIcon, ERR_NewObject));

   StepBack();
   return(ERR_Okay);
}

/******************************************************************************

###ACTION###
Name:  Clear
Short: Clears a view of all internal content and updates the display.
###END###

******************************************************************************/

static ERROR VIEW_Clear(struct View *Self, APTR Void)
{
   struct XMLTag **xmllist, *tag;
   struct ViewNode *node;
   BYTE activate;

   DPrintF("~Clear()","[View:%d]", Self->Head.UniqueID);

   if (Self->SelectedTag != -1) activate = TRUE;
   else activate = FALSE;

   Self->XPosition = 0;
   Self->YPosition = 0;
   Self->HighlightTag = -1;
   SetFields(Self, FID_SelectedTag|TLONG, -1, TAGEND); /* Use SetField in case the field is watched */
   Self->ActiveTag    = -1;

   /*** Free any loaded icons ***/

   if ((GetField(Self->XML, FID_Tags, FT_POINTER, &xmllist) IS ERR_Okay) AND (xmllist)) {
      for (tag=xmllist[0]; tag; tag=tag->Next) {
         node = tag->Private;
         if ((node->Icon) AND (!(node->Flags & NODE_CACHEICON))) acFree(node->Icon);
      }
   }

   acClear(Self->XML);

   ArrangeItems(Self);

   ActionMsg(AC_Draw, Self->DrawableID, NULL);

   if (activate) acActivate(Self);

   StepBack();
   return(ERR_Okay);
}

/*****************************************************************************
** Aciton: ClosingTag
*/

static ERROR VIEW_ClosingTag(struct View *Self, APTR Void)
{
   if (Self->SelectionIndex != -1) {
      SET_SelectionIndex(Self, &Self->SelectionIndex);
      Self->SelectionIndex = -1;
   }

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  DataChannel
Short: Items can be added to the view using data channels.

###DESCRIPTION###
<p>To add new items to a view, it is recommended that you use the data channel
system.  New items can be added as raw text or as a series of XML statements.</p>

<p>When DATA_TEXT data is sent to a view, it will be added as a single item to the
view list.  You will need to send multiple fragments of text data if you need
to more than one item using this method.</p>

<p>DATA_XML information is supported in an abstract fashion.  Each root tag that you
pass to the view will be interpreted as a new item.  Child items within each
root tag will be interpreted as extra information, to be used in certain
view types such as column mode (please refer to the <a href="#Columns">Columns</a> field for
further information on special tag types).  Special attributes may also be
set in each root tag for interpretation by the view object.  Currently the
"icon" attribute is available for defining the icon that should be displayed
against the item when icon graphics are enabled in the view.</p>

<p>The following example demonstrates the XML code that the
FileView class uses to add new files to view objects:</p>

<pre>
  &lt;file icon="filetypes/audio"&gt;audio.wav
    &lt;size sort="0002106"&gt;2106&lt;/size&gt;
    &lt;date&gt;20031118 14:10:32&lt;/date&gt;
  &lt;/file&gt;
</pre>

<p>The content of the root tag, in this case "audio.wav" is used as the item
text.  An icon has been specified to accompany the item.  The size and date
tags are special tag types that are specific to column mode and will
be ignored when not in use.</p>

###END###

*****************************************************************************/

static ERROR VIEW_DataChannel(struct View *Self, struct acDataChannel *Args)
{
   struct XMLTag **taglist, *tag;
   struct ViewNode *node;
   struct acDraw draw;
   STRING iconfile;
   LONG tagcount;
   ERROR error;
   UBYTE buffer[300];
   LONG keyvalue, keyflags, j;

   taglist = NULL;

   GetFields(Self->XML, FID_Tags|TPTR,      &taglist,
                        FID_TagCount|TLONG, &tagcount,
                        TAGEND);

   /* Pass XML information on the XML object.  Then refresh the display to
   ** include the new information.
   */

   if (!Args) return(ObjectError(Self, ERH_DataChannel, ERR_Args));

   if ((Args->DataType IS DATA_KEYPRESS) AND (Args->Version <= 1)) {
      for (j=0; j < Args->TotalEntries; j++) {
         if (((struct dcKeyEntry *)Args->Buffer)[j].Flags & KQ_PRESSED) {
            keyflags = ((struct dcKeyEntry *)Args->Buffer)[j].Flags;
            keyvalue = ((struct dcKeyEntry *)Args->Buffer)[j].Value;

            if (keyflags & KQ_CTRL) {

            }
            else {
               switch(keyvalue) {
                  case K_ENTER:
                     Action(AC_Activate, Self, NULL);
                     break;

                  case K_LEFT:
                     break;

                  case K_RIGHT:
                     break;

                  case K_DOWN:
                     if (!taglist) break;

                     if ((Self->Style IS VIEW_COLUMN) OR (Self->Style IS VIEW_LIST) OR (Self->Style IS VIEW_LONGLIST)) {
                        if (Self->ActiveTag IS -1) SelectItem(Self, taglist[0], FALSE);
                        else if (taglist[Self->ActiveTag]->Next) SelectItem(Self, taglist[Self->ActiveTag]->Next, FALSE);
                     }
                     break;

                  case K_UP:
                     if (!taglist) break;

                     if ((Self->Style IS VIEW_COLUMN) OR (Self->Style IS VIEW_LIST) OR (Self->Style IS VIEW_LONGLIST)) {
                        if (Self->ActiveTag IS -1) SelectItem(Self, taglist[0], FALSE);
                        else if (taglist[Self->ActiveTag]->Prev) SelectItem(Self, taglist[Self->ActiveTag]->Prev, FALSE);
                     }
                     break;
               }
            }
         }
      }
      return(ERR_Okay);
   }
   else if ((Args->DataType IS DATA_XML) AND (Args->Version <= 1)) {
      /*** Pass the data straight through to XML ***/
//return(ERR_Okay);

      if (Action(AC_DataChannel, Self->XML, Args) != ERR_Okay) {
         return(ObjectError(Self, ERH_DataChannel, ERR_Failed));
      }

      if ((GetFields(Self->XML, FID_Tags|TPTR, &taglist, TAGEND) != ERR_Okay) OR (!taglist)) {
         return(ObjectError(Self, ERH_DataChannel, ERR_GetField));
      }

      /* Load newly referenced icons.  Icons must be referenced
      ** in the format "group/iconname".  We will add the icons:
      ** assignment and determine the required icon size.
      */

      if (!(Self->Flags & VWF_NOICONS)) {
         for (tag=taglist[tagcount]; tag; tag=tag->Next) {
            if ((iconfile = GetXMLAttribute(tag, "icon"))) {
               SetIcon(Self, tag, iconfile);
            }
         }
      }

      /*** Set default colour for new items ***/

      for (tag=taglist[tagcount]; tag; tag=tag->Next) {
         node = tag->Private;
         node->FontRGB.Red   = Self->ItemRGB.Red;
         node->FontRGB.Green = Self->ItemRGB.Green;
         node->FontRGB.Blue  = Self->ItemRGB.Blue;
      }

      /*** Refresh the display ***/

      if (Self->Sort[0]) {
         error = SortItems(Self);
         ArrangeItems(Self);
         if (error IS ERR_Okay) ActionMsg(AC_Draw, Self->DrawableID, NULL);
         else {
            /* The list didn't need to be sorted, so just draw the new items.
            ** If a new column had to be added then we will need to redraw
            ** the entire view though.
            */

            for (tag=taglist[tagcount]; tag; tag=tag->Next) {
               node = tag->Private;
               if (node->Flags & NODE_NEWCOLUMN) {
                  ActionMsg(AC_Draw, Self->DrawableID, NULL);
                  break;
               }
            }

            if (!tag) {
               for (tag=taglist[tagcount]; tag; tag=tag->Next) {
                  DrawItem(Self, tag);
               }
            }
         }
      }
      else {
         ArrangeItems(Self);

         /*** Draw only the newly added items ***/

         for (tag=taglist[tagcount]; tag; tag=tag->Next) {
            node = tag->Private;
            if (node->Flags & NODE_NEWCOLUMN) {
               ActionMsg(AC_Draw, Self->DrawableID, NULL);
               break;
            }
         }

         if (!tag) {
            for (tag=taglist[tagcount]; tag; tag=tag->Next) {
               DrawItem(Self, tag);
            }
         }
      }
   }
   else if ((Args->DataType IS DATA_TEXT) AND (Args->Version <= 1)) {
      struct acDataChannel channel;

      /*** Pass the data through to XML ***/

      GetFields(Self->XML, FID_TagCount|TLONG, &tagcount, TAGEND);

      StrFormat(buffer, sizeof(buffer), "<item>%s</item>", Args->Buffer);

      channel.ObjectID = Self->Head.UniqueID;
      channel.DataType = DATA_XML;
      channel.Version  = 1;
      channel.Buffer   = buffer;
      for (channel.Size=0; buffer[channel.Size]; channel.Size++);
      channel.TotalEntries = 1;
      if (Action(AC_DataChannel, Self->XML, &channel) != ERR_Okay) {
         return(ERR_Failed);
      }

      if (GetFields(Self->XML, FID_Tags|TPTR, &taglist, TAGEND) != ERR_Okay) {
         return(ObjectError(Self, ERH_DataChannel, ERR_GetField));
      }

      /*** Set default colour for new items ***/

      for (tag=taglist[tagcount]; tag; tag=tag->Next) {
         node = tag->Private;
         node->FontRGB.Red   = Self->ItemRGB.Red;
         node->FontRGB.Green = Self->ItemRGB.Green;
         node->FontRGB.Blue  = Self->ItemRGB.Blue;
      }

      /*** Refresh the display ***/

      if (Self->Sort[0]) {
         error = SortItems(Self);
         ArrangeItems(Self);
         if (error IS ERR_Okay) ActionMsg(AC_Draw, Self->DrawableID, NULL);
         else {
            /*** The list didn't need to be sorted, so just draw the new items ***/
            for (tag=taglist[tagcount]; tag; tag=tag->Next) {
               node = tag->Private;
               draw.XCoord = Self->AreaX + node->XCoord;
               draw.YCoord = Self->AreaY + node->YCoord;
               draw.Width  = node->Width;
               if (node->Flags & NODE_NEWCOLUMN) draw.Height = 16000;
               else draw.Height = node->Height;
               ActionMsg(AC_Draw, Self->DrawableID, &draw);
            }
         }
      }
      else {
         ArrangeItems(Self);

         for (tag=taglist[tagcount]; tag; tag=tag->Next) {
            node = tag->Private;
            draw.XCoord = Self->AreaX + node->XCoord;
            draw.YCoord = Self->AreaY + node->YCoord;
            draw.Width  = node->Width;
            if (node->Flags & NODE_NEWCOLUMN) draw.Height = 16000;
            else draw.Height = node->Height;
            ActionMsg(AC_Draw, Self->DrawableID, &draw);
         }
      }

   }
   else DPrintF("@DataChannel:","[View:%d] Data type #%d not supported.", Self->Head.UniqueID, Args->DataType);

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name: Disable
Short: Disables the view.
###END###

*****************************************************************************/

static ERROR VIEW_Disable(struct View *Self, APTR Void)
{
   acDisableID(Self->DrawableID);
   return(ERR_Okay);
}

/*****************************************************************************
** View: Draw
*/

static ERROR VIEW_Draw(struct View *Self, struct acDraw *Args)
{
   ActionMsg(AC_Draw, Self->DrawableID, Args);
   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name: Enable
Short: Enables a view that has been disabled.
###END###

*****************************************************************************/

static ERROR VIEW_Enable(struct View *Self, APTR Void)
{
   acEnableID(Self->DrawableID);
   return(ERR_Okay);
}

/*****************************************************************************
** View: Free
*/

static ERROR VIEW_Free(struct View *Self, APTR Void)
{
   struct Column *col, *next;
   OBJECTPTR object;
   WORD i;

   for (i=0; i < ARRAYSIZE(Self->Cache); i++) {
      if (Self->Cache[i].Picture) acFree(Self->Cache[i].Picture);
   }

   if (Self->DrawableID) {
      if (AccessObject(Self->DrawableID, 5000, &object) IS ERR_Okay) {
         ActionTags(MT_SetRenderCallback, object, Self, NULL);
         UnsubscribeAction(object, NULL, GetUniqueID(Self));
         ReleaseObject(object);
      }
   }

   /*** Release the keyboard ***/

   if ((Self->InputID) AND (Self->KeyboardChannel IS TRUE)) {
      if (AccessObject(Self->InputID, 5000, &object) IS ERR_Okay) {
         UnsubscribeChannel(object, Self->Head.UniqueID);
         ReleaseObject(object);
      }
      Self->InputID = NULL;
   }

   if (Self->SelectedTags) {
      FreeMemory(Self->SelectedTags);
      Self->SelectedTags = NULL;
   }

   /*** Free column allocations ***/

   col = Self->Columns;
   while (col) {
      next = col->Next;
      FreeMemory(col);
      col = next;
   }

   if (Self->XML) {
      struct XMLTag **xmllist, *tag;
      struct ViewNode *node;

      /*** Free any loaded icons ***/

      if ((GetField(Self->XML, FID_Tags, FT_POINTER, &xmllist) IS ERR_Okay) AND (xmllist)) {
         for (tag=xmllist[0]; tag; tag=tag->Next) {
            node = tag->Private;
            if ((node->Icon) AND (!(node->Flags & NODE_CACHEICON))) acFree(node->Icon);
         }
      }

      acFree(Self->XML);
      Self->XML = NULL;
   }

   if (Self->Font) {
      acFree(Self->Font);
      Self->Font = NULL;
   }

   if (Self->ColumnString) {
      FreeMemory(Self->ColumnString);
      Self->ColumnString = NULL;
   }

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Focus
Short: Activates keyboard monitoring (by force).
###END###

*****************************************************************************/

static ERROR VIEW_Focus(struct View *Self, APTR Void)
{
   return(acFocusID(Self->DrawableID));
}

/*****************************************************************************

###ACTION###
Name:  GetUnlistedField
Short: Special field types are available via unlisted field support.

###DESCRIPTION###
<p>The View class supports the following field types via the unlisted
field mechanism:</p>

<blockquote>
<p><i>Selection(Index,Attribute)</i> Returns the content of the selection at
the requested Index (the Index value may not exceed the value in the TotalSelected
field).  The Attribute parameter is optional.  It should be specified if there
is a certain XML attribute that you want to read instead of the default item
content.</p>

<p><i>Active(Attribute)</i> Returns the content of the most recently active
item.  The most recently active item is defined as the one that the
user has most recently clicked on or activated.  The item does not
necessarily have to be selected.</p>
</blockquote>

###END###

*****************************************************************************/

static ERROR VIEW_GetUnlistedField(struct View *Self, struct acGetUnlistedField *Args)
{
   struct XMLTag *tag, **xmllist;
   LONG index, tagcount;

   if (!Args) return(ObjectError((OBJECTPTR)Self, ERH_GetUnlistedField, ERR_Args));

   if ((!Args->Field) OR (!Args->Buffer) OR (Args->Size < 1)) {
      return(ObjectError((OBJECTPTR)Self, ERH_GetUnlistedField, ERR_Args));
   }

   if (StrCompare("active", Args->Field, 0, NULL) IS ERR_Okay) {
      struct XMLTag *tag;
      UBYTE attrib[60];
      WORD i, j;

      if (Self->ActiveTag IS -1) return(ERR_NoData);

      if ((GetFields(Self->XML, FID_Tags|TPTR,      &xmllist,
                                FID_TagCount|TLONG, &tagcount,
                                TAGEND) != ERR_Okay) OR (!xmllist)) return(ERR_NoData);

      if (Self->ActiveTag >= tagcount) return(ERR_NoData);
      if (!(tag = xmllist[Self->ActiveTag])) return(ERR_NoData);

      if (Args->Field[6] IS '(') {
         for (j=0,i=7; (j < sizeof(attrib)-1) AND (Args->Field[i]) AND (Args->Field[i] != ')'); j++) attrib[j] = Args->Field[i++];
         attrib[j] = 0;
      }
      else attrib[0] = 0;

      /*** Write the item to the field buffer ***/

      Args->Buffer[0] = 0;
      if (attrib[0]) {
         /*** Scan attributes in the master tag ***/

         for (i=0; i < tag->TotalAttributes; i++) {
            if (StrCompare(tag->Attributes[i].Name, attrib, 0, STR_MATCHLENGTH) IS ERR_Okay) {
               StrCopy(tag->Attributes[i].Value, Args->Buffer, Args->Size);
               break;
            }
         }

         /*** If we didn't find anything, scan child tags (column names) ***/

         if (!Args->Buffer[0]) {
            if ((tag = tag->Child)) {
               while (tag) {
                  if (StrCompare(tag->Attributes[0].Name, attrib, 0, STR_MATCHLENGTH) IS ERR_Okay) {
                     ActionTags(MT_GetXMLContent, Self->XML, tag->Index, Args->Buffer, Args->Size);
                     break;
                  }
                  tag = tag->Next;
               }
            }
         }
      }
      else ActionTags(MT_GetXMLContent, Self->XML, tag->Index, Args->Buffer, Args->Size);

      return(ERR_Okay);
   }
   else if (StrCompare("selection(", Args->Field, 0, NULL) IS ERR_Okay) {
      struct ViewNode *node;
      UBYTE attrib[60];
      WORD i, j;

      Args->Buffer[0] = '-';
      Args->Buffer[1] = '1';
      Args->Buffer[2] = 0;

      index = StrToInt(Args->Field) + 1;
      if (index < 1) return(ERR_Okay);

      for (i=0; Args->Field[i] AND (Args->Field[i] != ','); i++);
      if (Args->Field[i] IS ',') {
         i++;
         for (j=0; (j < sizeof(attrib)-1) AND (Args->Field[i]) AND (Args->Field[i] != ')'); j++) attrib[j] = Args->Field[i++];
         attrib[j] = 0;
      }
      else attrib[0] = 0;

      /*** Find the requested item ***/

      if ((GetField(Self->XML, FID_Tags, FT_POINTER, &xmllist) != ERR_Okay) OR (!xmllist)) {
         return(ERR_Okay);
      }

      i = 0;
      for (tag=xmllist[0]; tag; tag = tag->Next) {
         node = tag->Private;
         if (node->Flags & NODE_SELECTED) {
            if (--index < 1) break;
         }
      }

      if (!tag) return(ERR_Okay);

      /*** Write the item to the field buffer ***/

      Args->Buffer[0] = 0;
      if (attrib[0]) {
         for (i=0; i < tag->TotalAttributes; i++) {
             if (StrCompare(tag->Attributes[i].Name, attrib, 0, STR_MATCHLENGTH) IS ERR_Okay) {
                StrCopy(tag->Attributes[i].Value, Args->Buffer, Args->Size);
                break;
             }
         }

         /*** If we didn't find anything, scan child tags (column names) ***/

         if (!Args->Buffer[0]) {
            if ((tag = tag->Child)) {
               while (tag) {
                  if (StrCompare(tag->Attributes[0].Name, attrib, 0, STR_MATCHLENGTH) IS ERR_Okay) {
                     ActionTags(MT_GetXMLContent, Self->XML, tag->Index, Args->Buffer, Args->Size);
                     break;
                  }
                  tag = tag->Next;
               }
            }
         }
      }
      else ActionTags(MT_GetXMLContent, Self->XML, tag->Index, Args->Buffer, Args->Size);

      return(ERR_Okay);
   }

   return(ERR_NoSupport);
}

/*****************************************************************************

###ACTION###
Name:  Hide
Short: Hides the view from the display.
###END###

*****************************************************************************/

static ERROR VIEW_Hide(struct View *Self, APTR Void)
{
   if (!(Self->Flags & VWF_INVISIBLE)) {
      Self->Flags |= VWF_INVISIBLE;
      ActionMsg(AC_Draw, Self->DrawableID, NULL);
   }
   return(ERR_Okay);
}

/*****************************************************************************
** View: Init()
*/

static ERROR VIEW_Init(struct View *Self, APTR Void)
{
   struct Variable var;
   struct Render *drawable;
   OBJECTID owner_id;
   OBJECTPTR object;

   /*** Find the drawable object that we are associated with ***/

   if (!Self->DrawableID) {
      owner_id = GetOwner(Self);
      while ((owner_id) AND (GetClassID(owner_id) != ID_RENDER)) {
         owner_id = GetOwnerID(owner_id);
      }
      if (!owner_id) return(ObjectError(Self, ERH_Init, ERR_UnsupportedOwner));
      else Self->DrawableID = owner_id;
   }

   if (!Self->FocusID) Self->FocusID = Self->DrawableID;

   /*** Setup keyboard ***/

   if (!Self->InputID) {
      if (FastFindObject("SystemKeyboard", NULL, &Self->InputID, 1, NULL) != ERR_Okay) {
         DPrintF("@Init:","[View] Warning: Could not find any SystemKeyboard objects.");
      }
   }

   /*** Initialise the XML object ***/

   if (acInit(Self->XML) != ERR_Okay) {
      return(ObjectError(Self, ERH_Init, ERR_Init));
   }

   /*** Initialise the font object ***/

   Self->Font->Flags |= FTF_CHARCLIP;
   Self->Font->WrapWidth = 8192;
   if (acInit(Self->Font) != ERR_Okay) {
      return(ObjectError(Self, ERH_Init, ERR_Init));
   }

   /* If dimension settings are missing (e.g. if it is impossible to
   ** determine width, height or a coordinate), then we set the missing
   ** fields to maximum possible values.
   */

   var.Type = FD_DOUBLE;
   var.Double = 0;
   if (!(Self->Dimensions & (DMF_FIXED_X|DMF_RELATIVE_X|DMF_FIXED_XOFFSET|DMF_RELATIVE_XOFFSET))) {
      SET_X(Self, &var);
   }

   if (!(Self->Dimensions & (DMF_FIXED_Y|DMF_RELATIVE_Y|DMF_FIXED_YOFFSET|DMF_RELATIVE_YOFFSET))) {
      SET_Y(Self, &var);
   }

   if (!(Self->Dimensions & (DMF_FIXED_WIDTH|DMF_RELATIVE_WIDTH|DMF_FIXED_XOFFSET|DMF_RELATIVE_XOFFSET))) {
      SET_XOffset(Self, &var);
   }

   if (!(Self->Dimensions & (DMF_FIXED_HEIGHT|DMF_RELATIVE_HEIGHT|DMF_FIXED_YOFFSET|DMF_RELATIVE_YOFFSET))) {
      SET_YOffset(Self, &var);
   }

   /*** Subscribe to our parent drawable ***/

   if (AccessObject(Self->DrawableID, 5000, (OBJECTPTR *)&drawable) IS ERR_Okay) {
      ActionTags(MT_SetRenderCallback, drawable, Self, &DrawView);

      SubscribeActionTags(drawable, Self,
         AC_Disable,
         AC_Enable,
         AC_Redimension,
         AC_UserClick,
         AC_UserClickRelease,
         AC_UserMovement,
         TAGEND);

      var.Type = FD_LARGE;
      GET_X(Self, &var); Self->AreaX = var.Large;
      GET_Y(Self, &var); Self->AreaY = var.Large;
      GET_Width(Self, &var);  Self->AreaWidth  = var.Large;
      GET_Height(Self, &var); Self->AreaHeight = var.Large;

      /*** Scan for scrollbars ***/

      if ((!Self->VScroll) OR (!Self->HScroll)) {
         LONG direction, count;
         OBJECTID scrollid;

         if ((count = TotalChildren(Self->DrawableID))) {
            struct ChildEntry list[count];

            if (ListChildren(Self->DrawableID, list, &count) IS ERR_Okay) {
               for (--count; count >= 0; count--) {
                  if (list[count].ClassID IS ID_SCROLLBAR) {
                     if (AccessObject(list[count].ObjectID, 3000, &object) IS ERR_Okay) {

                        GetField(object, FID_Direction, FT_LONG, &direction);

                        if ((direction IS SO_HORIZONTAL) AND (!Self->HScroll)) {
                           GetField(object, FID_Scroll, FT_LONG, &scrollid);
                           SET_HScroll(Self, &scrollid);
                        }
                        else if ((direction IS SO_VERTICAL) AND (!Self->VScroll)) {
                           GetField(object, FID_Scroll, FT_LONG, &scrollid);
                           SET_VScroll(Self, &scrollid);
                        }

                        ReleaseObject(object);
                     }
                  }
               }
            }
         }
      }

      ReleaseObject(drawable);
   }
   else return(ObjectError(Self, ERH_Init, ERR_AccessObject));

   DPrintF("Init:","[View] Focus notification based on object #%d.", Self->FocusID);

   if (AccessObject(Self->FocusID, 5000, &object) IS ERR_Okay) {
      SubscribeActionTags(object, Self,
         AC_Focus,
         AC_LostFocus,
         TAGEND);

      /* Test the focus object to see if it already has the focus and
      ** if so, activate keyboard monitoring.
      */

      if (object->ClassID IS ID_RENDER) {
         if (((struct Render *)object)->Flags & RNF_HASFOCUS) InitKeyboard(Self);
      }

      ReleaseObject(object);
   }

   ArrangeItems(Self);

   return(ERR_Okay);
}

/*****************************************************************************

###METHOD###
Name:  InvertSelection
Short: Inverts all currently selected items in the view.

###DESCRIPTION###
<p>This method will invert the selections in the view, so that selected items
become deselected and all other items are selected.  The view will be redrawn
as a result of calling this method.</p>

###END###

*****************************************************************************/

static ERROR VIEW_InvertSelection(struct View *Self, APTR Void)
{
   struct XMLTag **xmllist, *tag;
   struct ViewNode *node;

   if ((GetField(Self->XML, FID_Tags, FT_POINTER, &xmllist) IS ERR_Okay) AND (xmllist)) {
      for (tag=xmllist[0]; tag; tag=tag->Next) {
         node = tag->Private;
         node->Flags = node->Flags ^ NODE_SELECTED;
      }

      Self->ActiveTag    = -1;
      Self->HighlightTag = -1;

      /*** Select the first tag in the list ***/

      for (tag=xmllist[0]; tag; tag=tag->Next) {
         node = tag->Private;
         if (node->Flags & NODE_SELECTED) {
            SetFields(Self, FID_SelectedTag|TLONG, tag->Index, TAGEND);
            break;
         }
      }

      if (!tag) SetFields(Self, FID_SelectedTag|TLONG, -1, TAGEND);

      ActionMsg(AC_Draw, Self->DrawableID, NULL);
   }

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  LostFocus
Short: Deactivates keyboard monitoring (by force).
###END###

*****************************************************************************/

static ERROR VIEW_LostFocus(struct View *Self, APTR Void)
{
   OBJECTPTR input;

   DPrintF("~LostFocus()","[View:%d]", Self->Head.UniqueID);

   if (Self->KeyboardChannel IS TRUE) {
      if (AccessObject(Self->InputID, 5000, &input) IS ERR_Okay) {
         UnsubscribeChannel(input, Self->Head.UniqueID);
         Self->KeyboardChannel = FALSE;
         ReleaseObject(input);
      }
      else ObjectError(Self, ERH_ActionNotify, ERR_AccessObject);
   }

   StepBack();
   return(ERR_Okay);
}

/*****************************************************************************
** View: NewObject()
*/

static ERROR VIEW_NewObject(struct View *Self, APTR Void)
{
   OBJECTID varid;
   OBJECTPTR var;
   WORD i;

   if (NewPrivateObject(ID_XML, NF_CHILD, &Self->XML) IS ERR_Okay) {
      SetFields(Self->XML, FID_PrivateDataSize|TLONG, sizeof(struct ViewNode),
                           FID_Flags|TLONG, XMF_STRIPHEADERS,
                           TAGEND);

      if (NewPrivateObject(ID_FONT, NF_CHILD, &Self->Font) IS ERR_Okay) {
         SetField(Self->Font, FID_Face, FT_STRING, "sans serif:9");

         if (FastFindObject("gl_fonts", ID_VARIABLE, &varid, 1, NULL) IS ERR_Okay) {
            if (AccessObject(varid, 2000, &var) IS ERR_Okay) {
               UBYTE buffer[80];

               if (GetFieldVariable(var, "face", buffer, sizeof(buffer)) IS ERR_Okay) {
                  SetField(Self->Font, FID_Face, FT_STRING, buffer);
               }

               ReleaseObject(var);
            }
         }

         if (AllocMemory(sizeof(struct Column), MEM_DATA, &Self->Columns, NULL) IS ERR_Okay) {
            StrCopy("Default", Self->Columns->Name, sizeof(Self->Columns->Name));
            StrCopy("Default", Self->Columns->Text, sizeof(Self->Columns->Text));
            Self->Columns->Type  = CT_VARIANT;
            Self->Columns->Width = 160;
            Self->Columns->Sort  = SORT_UNSORTED;

            StrCopy(glDateFormat, Self->DateFormat, sizeof(Self->DateFormat));

            for (i=1; i < ARRAYSIZE(Self->Sort); i++) Self->Sort[i] = 0;

            Self->Style        = VIEW_COLUMN;
            Self->MaxItemWidth = 170;
            Self->HSpacing     = 10;
            Self->VSpacing     = 2;
            Self->HighlightRGB.Red   = 255;
            Self->HighlightRGB.Green = 0;
            Self->HighlightRGB.Blue  = 0;
            Self->SelectRGB.Red   = 230; //220,220,245
            Self->SelectRGB.Green = 230;
            Self->SelectRGB.Blue  = 255;
            Self->HighlightTag    = -1;
            Self->SelectionIndex  = -1;
            Self->ActiveTag       = -1;
            Self->SelectedTag     = -1;

            Self->ColThickness        = 2;
            Self->ColBackground.Red   = 210;
            Self->ColBackground.Green = 210;
            Self->ColBackground.Blue  = 210;
            Self->ColHighlight.Red    = 255;
            Self->ColHighlight.Green  = 255;
            Self->ColHighlight.Blue   = 255;
            Self->ColShadow.Red       = 0;
            Self->ColShadow.Green     = 0;
            Self->ColShadow.Blue      = 0;
            Self->HairlineRGB.Red     = 200;
            Self->HairlineRGB.Green   = 200;
            Self->HairlineRGB.Blue    = 200;

            return(ERR_Okay);
         }
         else return(ERR_AllocMemory);
      }
      else return(ERR_NewObject);
   }
   else return(ERR_NewObject);
}

/*****************************************************************************

###ACTION###
Name:  Refresh
Short: Refreshes the display.

###DESCRIPTION###
<p>Calling the Refresh action will update the view so that its display
reflects the most current data in its XML definition.  A request for a
complete redraw of the view will be posted to the view's drawable.</p>

<p>Use of the Refresh action is most useful if you directly update the XML
content of the view without its knowledge.</p>

###END###

*****************************************************************************/

static ERROR VIEW_Refresh(struct View *Self, APTR Void)
{
   STRING iconfile;
   struct XMLTag *tag, **taglist;
   struct ViewNode *node;
   LONG tagcount;

   if ((GetFields(Self->XML, FID_Tags|TPTR,      &taglist,
                             FID_TagCount|TLONG, &tagcount,
                             TAGEND) IS ERR_Okay) AND (taglist)) {

      if (Self->HighlightTag >= tagcount) Self->HighlightTag = -1;
      if (Self->ActiveTag >= tagcount)    Self->ActiveTag    = -1;
      if (Self->SelectedTag >= tagcount)  SetFields(Self, FID_SelectedTag|TLONG, -1, TAGEND);

      /*** Check XML for icons that have been described but not loaded in yet ***/

      if (!(Self->Flags & VWF_NOICONS)) {
         for (tag=taglist[0]; tag; tag=tag->Next) {
            /*** Only load icons that have not been loaded yet ***/
            node = tag->Private;
            if ((!node->Icon) AND ((iconfile = GetXMLAttribute(tag, "icon")))) {
               SetIcon(Self, tag, iconfile);
            }
         }
      }
   }

   ArrangeItems(Self);
   ActionMsg(AC_Draw, Self->DrawableID, NULL);
   return(ERR_Okay);
}

/******************************************************************************

###METHOD###
Name:    RemoveViewItem
Synonym: RemoveItem
Short:   Removes an item from the view.

###DESCRIPTION###
<p>Use the RemoveViewItem() method to remove an item from the view.  This method is
provided mostly for script usage, as it can be faster to talk to the XML
object directly and use the Refresh action than to make multiple calls to RemoveViewItem().</p>

<p>In order to call RemoveViewItem() successfully, you should have an awareness
of the XML tag structure that you have used for creating the view items. You
will need to pass the ID of the item that you want to update (thus you must
have used the 'id' parameter when constructing the original XML tags), the
name of tag to change (NULL is acceptable for the default) and the attribute
to set (NULL is acceptable to change the XML content that represents the
item data).  The new string to be set is defined in the Value parameter.</p>

<p>If you update the content of an item by passing a NULL Attrib value, then
a completely redraw of the view will occur so that the display accurately
reflects the view data.</p>

###INPUT###
LONG/ID: The ID of the item that you want to remove, or -1 if you want to use a TagIndex.
LONG/TagIndex: The index of the XML tag that you want to remove (ID must be -1).  Ignored if the ID is set to any other value.

###RESULT###
ERR_Okay: Operation successful.
ERR_Args: The Icon argument was not specified.

###END###

******************************************************************************/

static ERROR VIEW_RemoveItem(struct View *Self, struct mtRemoveViewItem *Args)
{
   struct ViewNode *node;
   struct XMLTag **taglist, *tag;
   LONG i, tagcount, tagindex;

   if (!Args) return(ERR_Args);

   if ((GetFields(Self->XML, FID_Tags|TPTR,      &taglist,
                             FID_TagCount|TLONG, &tagcount,
                             TAGEND) IS ERR_Okay) AND (taglist)) {

      tag = NULL;
      if (Args->ID != -1) {
         /*** Find the root tag that we need to set ***/

         for (tag=*taglist; (tag); tag=tag->Next) {
            for (i=1; i < tag->TotalAttributes; i++) {
               if ((tag->Attributes[i].Name) AND (lcase(tag->Attributes[i].Name[0]) IS 'i') AND
                   (lcase(tag->Attributes[i].Name[1]) IS 'd') AND (!tag->Attributes[i].Name[2])) {
                  if (Args->ID IS StrToInt(tag->Attributes[i].Value)) goto found;
               }
            }
         }
      }
      else if ((Args->TagIndex >= 0) AND (Args->TagIndex < tagcount)) {
         tag = taglist[Args->TagIndex];
      }
      else return(ObjectError(Self, MT_RemoveViewItem, ERR_OutOfRange));
found:
      if (tag) {
         BYTE activate;

         tagindex = tag->Index;

         /* Check if the tag is currently selected.  If so, we'll need to send
         ** a reactivation message to children monitoring the view.
         */

         node = tag->Private;
         activate = (node->Flags & NODE_SELECTED) ? TRUE : FALSE;

         /*** Remove the tag and associated resources from the system ***/

         if ((node = tag->Private) AND (node->Icon) AND (!(node->Flags & NODE_CACHEICON))) acFree(node->Icon);
         ActionTags(MT_RemoveXMLTag, Self->XML, tagindex, 1);

         /*** Fix selection indexes ***/

         if (Self->HighlightTag IS tagindex) Self->HighlightTag = -1;
         if (Self->ActiveTag IS tagindex)    Self->ActiveTag    = -1;
         if (Self->SelectedTag IS tagindex)  SetFields(Self, FID_SelectedTag|TLONG, -1, TAGEND);

         ArrangeItems(Self);
         ActionMsg(AC_Draw, Self->DrawableID, NULL);

         /*** Send an Activate notification if necessary ***/

         if (activate) {
            DPrintF("RemoveItem:","[View] Reactivating due to deleted selected item.");
            acActivate(Self);
         }
      }
      else return(ObjectError(Self, MT_RemoveViewItem, ERR_Search));
   }

   return(ERR_Okay);
}

/******************************************************************************

###METHOD###
Name:    RemoveViewTag
Synonym: RemoveTag
Short:   Removes an XML tag without updating the view (for optimisation only).

###DESCRIPTION###
<p>Use the RemoveViewTag() method to delete an XML tag without updating the
view display.  This method is recommended for optimisation purposes only, and
serves as a replacement for the XML version of this method, RemoveXMLTag().
By using this method you can correctly delete an XML tag without having
adverse affects on the view (e.g. if the tag is associated with an icon, the
icon will be deallocated from the system).</p>

###INPUT###
LONG/TagIndex: The index of the XML tag that you want to remove.

###RESULT###
ERR_Okay: Operation successful.
ERR_Args: The Icon argument was not specified.

###END###

******************************************************************************/

static ERROR VIEW_RemoveTag(struct View *Self, struct mtRemoveViewTag *Args)
{
   struct ViewNode *node;
   struct XMLTag **taglist, *tag;
   LONG tagcount, tagindex;

   if (!Args) return(ERR_Args);

   if ((GetFields(Self->XML, FID_Tags|TPTR,      &taglist,
                             FID_TagCount|TLONG, &tagcount,
                             TAGEND) IS ERR_Okay) AND (taglist)) {

      if ((Args->TagIndex >= 0) AND (Args->TagIndex < tagcount)) {
         tag = taglist[Args->TagIndex];
      }
      else return(ObjectError(Self, MT_RemoveViewTag, ERR_OutOfRange));

      tagindex = tag->Index;

      /*** Remove the tag and associated resources from the system ***/

      if ((node = tag->Private) AND (node->Icon) AND (!(node->Flags & NODE_CACHEICON))) acFree(node->Icon);
      ActionTags(MT_RemoveXMLTag, Self->XML, tag->Index, 1);

      /*** Fix selection indexes ***/

      if (Self->HighlightTag IS tagindex) Self->HighlightTag = -1;
      if (Self->ActiveTag IS tagindex)    Self->ActiveTag    = -1;
      if (Self->SelectedTag IS tagindex)  SetFields(Self, FID_SelectedTag|TLONG, -1, TAGEND);

      /* Note that this is a technical routine that does not run a cleanup
      ** sub-routine for item rearrangement and notifications.
      */
   }

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  ScrollToPoint
Short: Scrolls the graphical content of a view.
###END###

*****************************************************************************/

static ERROR VIEW_ScrollToPoint(struct View *Self, struct acScrollToPoint *Args)
{
   struct mtMoveContent move;
   struct acDraw draw;
   OBJECTPTR drawable;
   LONG oldx, oldy, xchange, ychange, x, y;

   if (!Args) return(ObjectError(Self, ERH_ScrollToPoint, ERR_Args));

   if ((Args->XCoord IS Self->XPosition) AND (Args->YCoord IS Self->YPosition)) return(ERR_Okay);

   if (AccessObject(Self->DrawableID, 5000, &drawable) IS ERR_Okay) {
      if (Args->Flags & STF_X) x = -Args->XCoord;
      else x = Self->XPosition;

      if (Args->Flags & STF_Y) y = -Args->YCoord;
      else y = Self->YPosition;

      oldx = Self->XPosition;
      oldy = Self->YPosition;

      Self->XPosition = x;
      Self->YPosition = y;

      xchange = x - oldx;
      ychange = y - oldy;

      if (Self->Flags & VWF_PRESERVEBKGD) {
         draw.XCoord = Self->AreaX;
         draw.YCoord = Self->AreaY;
         draw.Width  = Self->AreaWidth;
         draw.Height = Self->AreaHeight;
         ActionMsg(AC_Draw, Self->DrawableID, &draw);
      }
      else if ((xchange >= Self->AreaWidth-10) OR (ychange >= Self->AreaHeight-10) OR
          (xchange <= -Self->AreaWidth+10) OR (ychange <= -Self->AreaHeight+10)) {
         draw.XCoord = Self->AreaX;
         draw.YCoord = Self->AreaY;
         draw.Width  = Self->AreaWidth;
         draw.Height = Self->AreaHeight;
         ActionMsg(AC_Draw, Self->DrawableID, &draw);
      }
      else {
         /*** Set horizontal movement parameters ***/

         if (x < oldx) {
            move.XCoord = Self->AreaX - xchange;
            move.Width  = Self->AreaWidth;
            move.XDest  = Self->AreaX;
         }
         else {
            move.XCoord = Self->AreaX - xchange;
            move.Width  = Self->AreaWidth;
            move.XDest  = Self->AreaX;
         }

         /*** Set vertical movement parameters ***/

         if (y < oldy) {
            move.YCoord = Self->AreaY - ychange;
            move.Height = Self->AreaHeight;
            move.YDest  = Self->AreaY;
         }
         else {
            move.YCoord = Self->AreaY - ychange;
            move.Height = Self->AreaHeight;
            move.YDest  = Self->AreaY;
         }

         move.ClipLeft   = Self->AreaX;
         move.ClipTop    = Self->AreaY;
         move.ClipRight  = Self->AreaX + Self->AreaWidth;
         move.ClipBottom = Self->AreaY + Self->AreaHeight;
         move.Flags      = MCF_REDRAW_UNDERLYING | ((Self->Style IS VIEW_COLUMN) ? MCF_NO_EXPOSE : 0);

         /* Do not move the column headers in column mode (i.e. we don't
         ** want to screw up column's vertical positioning)
         */

         if (Self->Style IS VIEW_COLUMN) move.ClipTop += Self->ColumnHeight;

         Action(MT_MoveContent, drawable, &move);

         /* Now we can move the column headers if we're using them */

         if (Self->Style IS VIEW_COLUMN) {
            move.ClipTop -= Self->ColumnHeight;
            move.YCoord = Self->AreaY;
            move.Height = Self->ColumnHeight;
            move.YDest  = Self->AreaY;
            Action(MT_MoveContent, drawable, &move);

            ExposeDrawable(Self->DrawableID, Self->AreaX, Self->AreaY, Self->AreaWidth, Self->AreaHeight, EXF_REDRAW_VOLATILE); //EXF_CHILDREN|EXF_REDRAW_VOLATILE_OVERLAP);
         }
      }

      ReleaseObject(drawable);
   }

   return(ERR_Okay);
}

/*****************************************************************************

###METHOD###
Name:  SelectAll
Short: Selects all items in the view for the user.

###DESCRIPTION###
<p>This method will select all items in the view, as if the user had selected
them himself.  The view will be redrawn as a result of calling this method.</p>

###END###

*****************************************************************************/

static ERROR VIEW_SelectAll(struct View *Self, APTR Void)
{
   struct XMLTag **xmllist, *tag;
   struct ViewNode *node;

   if ((GetField(Self->XML, FID_Tags, FT_POINTER, &xmllist) IS ERR_Okay) AND (xmllist)) {
      for (tag=xmllist[0]; tag; tag=tag->Next) {
         node = tag->Private;
         node->Flags |= NODE_SELECTED;
      }

      Self->ActiveTag    = 0;
      Self->HighlightTag = -1;
      SetFields(Self, FID_SelectedTag|TLONG, 0, TAGEND);

      ActionMsg(AC_Draw, Self->DrawableID, NULL);
   }

   return(ERR_Okay);
}

/*****************************************************************************

###METHOD###
Name:  SelectItem
Short: Manually select items for the user.

###DESCRIPTION###
<p>The SelectItem method is used to manually select items in the view for the
benefit of the user.  To select an item you need to indicate the name of the
XML tag that you want to scan for, and the item string that will be used for
finding the correct item.</p>

<p>To illustrate correct usage of this method, imagine that you have filled
the view with a series of tags similar to the following one:</p>

<pre>
  &lt;file icon="filetypes/audio"&gt;audio.wav
    &lt;size sort="0002106"&gt;2106&lt;/size&gt;
    &lt;date&gt;20031118 14:10:32&lt;/date&gt;
  &lt;/file&gt;
</pre>

<p>To select the first audio item we could use a Tag of 'file' and an item
of 'audio.wav'.  If we want to make the selection based on the icon, we
would use a Tag of 'file:icon' and an item of 'filetypes/audio'.  Note the
use of the colon character to indicate the name of the tag attribute in this
case.  We could also scan for the size and date tags simply by specifying those
names in the Tag field.</p>

<p>In the event that multiple tags could match the given parameters, only the
first tag will be selected by the selection routine.  If there is a high probability
that the data strings are repeated, you should implement a unique ID for each tag
and use that for the selection of specific tags.</p>

<p>Note that selections are inclusive with other selections when in
MULTISELECT mode. Use the Clear action first if you want the selection to be
the only selected item in the list.</p>

###INPUT###
STRING/Tag: The name of the tag that you want to scan for in each item.  If not set, the default tag will be scanned.
STRING/Item: The string that you want to match against for item selection.

###RESULT###
ERR_Okay: The item was selected successfully.
ERR_Args: Invalid arguments were specified.

###END###

*****************************************************************************/

static ERROR VIEW_SelectItem(struct View *Self, struct mtSelectItem *Args)
{
   struct XMLTag *current, *tag, **taglist;
   UBYTE buffer[300], tagname[50], attrib[50];
   WORD i;

   if ((!Args) OR (!Args->Item)) return(ObjectError(Self, MT_SelectItem, ERR_Args));

   DPrintF("SelectItem()","Tag: %s, Item: %s", Args->Tag, Args->Item);

   if ((GetFields(Self->XML, FID_Tags|TPTR, &taglist, TAGEND) IS ERR_Okay) AND (taglist)) {
      tagname[0] = 0;
      attrib[0]  = 0;
      if ((Args->Tag) AND (Args->Tag[0])) {
         for (i=0; (Args->Tag[i]) AND (Args->Tag[i] != ':') AND (i < sizeof(tag)); i++) tagname[i] = Args->Tag[i];
         tagname[i] = 0;

         if (Args->Tag[i] IS ':') StrCopy(Args->Tag+i+1, attrib, sizeof(attrib));
      }

      for (current=*taglist; current; current=current->Next) {
         tag = current;
         if (tagname[0]) {
            if (StrCompare(tagname, tag->Attributes[0].Name, 0, STR_MATCHLENGTH) != ERR_Okay) {
               if ((tag = tag->Child)) {
                  while (tag) {
                     if (StrCompare(tagname, tag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) break;
                     tag = tag->Next;
                  }
               }
               if (!tag) continue;
            }
         }

         if (attrib[0]) {
            /*** Scan for the requested attribute ***/

            for (i=1; i < tag->TotalAttributes; i++) {
               if (StrCompare(attrib, tag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
                  /*** Check for a match ***/

                  if (StrCompare(Args->Item, tag->Attributes[i].Value, 0, STR_MATCHLENGTH) IS ERR_Okay) {
                     SelectItem(Self, tag, TRUE);
                     return(ERR_Okay);
                  }
               }
            }
         }
         else if (ActionTags(MT_GetXMLContent, Self->XML, tag->Index, buffer, sizeof(buffer)) IS ERR_Okay) {
            if (StrCompare(Args->Item, buffer, 0, NULL) IS ERR_Okay) {
               SelectItem(Self, tag, TRUE);
               return(ERR_Okay);
            }
         }
      }
   }

   DPrintF("@SelectItem:","[View] Unable to find item \"%s\"", tagname);
   return(ERR_Search);
}

/*****************************************************************************

###METHOD###
Name:  SelectNone
Short: Deselects all currently selected items.

###DESCRIPTION###
<p>This method will deselect all items in the view, as if the user had deselected
them himself.  The view will be redrawn as a result of calling this method.</p>

###END###

*****************************************************************************/

static ERROR VIEW_SelectNone(struct View *Self, APTR Void)
{
   struct XMLTag **xmllist, *tag;
   struct ViewNode *node;

   if ((GetField(Self->XML, FID_Tags, FT_POINTER, &xmllist) IS ERR_Okay) AND (xmllist)) {
      for (tag=xmllist[0]; tag; tag=tag->Next) {
         node = tag->Private;
         node->Flags &= ~NODE_SELECTED;
      }

      Self->ActiveTag    = -1;
      Self->HighlightTag = -1;
      SetFields(Self, FID_SelectedTag|TLONG, -1, TAGEND);

      ActionMsg(AC_Draw, Self->DrawableID, NULL);
   }

   return(ERR_Okay);
}

/*****************************************************************************

###METHOD###
Name:    SetViewItem
Synonym: SetItem
Short:   Changes the attributes of any item in the view.

###DESCRIPTION###
<p>Use the SetViewItem() method to change the attributes of items that are in
in the view.  This method is provided mostly for script usage, as it can be faster
to talk to the XML object directly and use the Refresh action than to make multiple
calls to this method.</p>

<p>In order to call SetViewItem() successfully, you should have an awareness
of the XML tag structure that you have used for creating the view items. You
will need to pass the ID of the item that you want to update (thus you must
have used the 'id' parameter when constructing the original XML tags), the
name of tag to change (NULL is acceptable for the default) and the attribute
to set (NULL is acceptable to change the XML content that represents the
item data).  The new string to be set is defined in the Value parameter.</p>

<p>If you update the content of an item by passing a NULL Attrib value, then
a completely redraw of the view will occur so that the display accurately
reflects the view data.</p>

###INPUT###
LONG/ID: The ID of the item that you want to set.  Set to -1 if you want to use a tag index.
LONG/TagIndex: The index of the XML tag that you want to set.  The ID must be set to -1.  Ignored if the ID is set to any other value.
STRING/Tag: The name of the child tag that you want to set within the item.  Set to NULL for the default tag.
STRING/Attrib: The name of the attribute that you want to set in the XML tag.  May be NULL to set the tag's content.
STRING/Value: The value that you want to set against the attribute.

###RESULT###
ERR_Okay: The item was set successfully.
ERR_Args: Invalid arguments were specified.
ERR_Search: The index, tag and/or attrib values did not lead to a match.

###END###

*****************************************************************************/

static ERROR VIEW_SetItem(struct View *Self, struct mtSetViewItem *Args)
{
   struct XMLTag **taglist, *tag;
   LONG index, tagcount;
   WORD i;

   if (!Args) return(ObjectError(Self, MT_SetViewItem, ERR_Args));

   DPrintF("~SetItem()","[View:%d] ID: %d, Tag: %s, Attrib: %s, Value: %s", Self->Head.UniqueID, Args->ID, Args->Tag, Args->Attrib, Args->Value);

   if ((GetFields(Self->XML, FID_Tags|TPTR,      &taglist,
                             FID_TagCount|TLONG, &tagcount,
                             TAGEND) IS ERR_Okay) AND (taglist)) {
      /*** Find the root tag that we need to set ***/

      if (Args->ID != -1) {
         for (tag=*taglist; tag; tag=tag->Next) {
            for (i=1; i < tag->TotalAttributes; i++) {
               if ((tag->Attributes[i].Name) AND (lcase(tag->Attributes[i].Name[0]) IS 'i') AND
                   (lcase(tag->Attributes[i].Name[1]) IS 'd') AND (!tag->Attributes[i].Name[2])) {
                  if (Args->ID IS StrToInt(tag->Attributes[i].Value)) goto found;
               }
            }
         }
      }
      else if ((Args->TagIndex >= 0) AND (Args->TagIndex < tagcount)) {
         tag = taglist[Args->TagIndex];
      }
      else return(StepObjectError(Self, MT_SetViewItem, ERR_OutOfRange));

found:
      if (!tag) return(StepObjectError(Self, MT_SetViewItem, ERR_Search));

      /*** Scan for the correct tag within the discovered area ***/

      if ((Args->Tag) AND (Args->Tag[0])) {
         if (StrCompare(Args->Tag, tag->Attributes[0].Name, 0, STR_MATCHLENGTH) != ERR_Okay) {
            for (tag=tag->Child; tag; tag=tag->Next) {
               if (StrCompare(Args->Tag, tag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) break;
            }
         }

         if (!tag) return(StepObjectError(Self, MT_SetViewItem, ERR_Search));
      }

      if ((Args->Attrib) AND (Args->Attrib[0])) {
         /*** Update an attribute.  There is no need to preform a redraw in this case ***/

         for (index=0; index < tag->TotalAttributes; index++) {
            if (StrCompare(Args->Attrib, tag->Attributes[index].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {

               if (StrCompare("icon", Args->Attrib, 0, STR_MATCHLENGTH) IS ERR_Okay) {
                  /* An icon change has been made.  We need to delete the existing icon
                  ** and load the one that has been specified.
                  */

                  SetIcon(Self, tag, Args->Value);
                  DrawItem(Self, tag);
               }

               ActionTags(MT_SetXMLAttribute, Self->XML, tag->Index, index, NULL, Args->Value);

               StepBack();
               return(ERR_Okay);
            }
         }
      }
      else {
         /*** Update content ***/
         if (tag->Child) {
            ActionTags(MT_SetXMLAttribute, Self->XML, tag->Child->Index, NULL, NULL, Args->Value);

            /* Items need to be rearranged (due to issues like extending
            ** node widths, sorting etc).
            */

            ArrangeItems(Self);
            ActionMsg(AC_Draw, Self->DrawableID, NULL);
         }

         StepBack();
         return(ERR_Okay);
      }
   }

   return(StepObjectError(Self, MT_SetViewItem, ERR_Search));
}

/*****************************************************************************

###ACTION###
Name:  Show
Short: Redisplays the view if it has been hidden.
###END###

*****************************************************************************/

static ERROR VIEW_Show(struct View *Self, APTR Void)
{
   if (Self->Flags & VWF_INVISIBLE) {
      Self->Flags &= ~VWF_INVISIBLE;
      ActionMsg(AC_Draw, Self->DrawableID, NULL);
   }
   return(ERR_Okay);
}

/*****************************************************************************

###METHOD###
Name:  SortColumn
Short: Sorts the view by column.

###DESCRIPTION###
<p>The SortColumn routine will sort the View on the column name that
you indicate.  The sort will be ascending by default, which you can reverse
by setting the Descending parameter to TRUE.  The View will automatically
redraw itself to reflect the newly sorted content.</p>

<p>This method will fail if the referenced column does not exist.  Use a
Column value of NULL if you want to sort on item content by default
(useful if you have not defined any columns for the view).</p>

###INPUT###
STRING/Column: The name of the column you want to sort by.
LONG/Descending: Set to TRUE if the sort should be descending.

###RESULT###
ERR_Okay: The column was sorted successfully.
ERR_Args: Invalid arguments were specified.
ERR_Search: The referenced column could not be found.

###END###

*****************************************************************************/

static ERROR VIEW_SortColumn(struct View *Self, struct mtSortColumn *Args)
{
   struct mtSortColumnIndex sort;
   struct Column *col;
   WORD colindex;

   if (!Args) return(ERR_Args);

   if (!Args->Column) {
      sort.Column     = 0;
      sort.Descending = Args->Descending;
      return(VIEW_SortColumnIndex(Self, &sort));
   }

   /*** Find the column that has been referenced ***/

   for (col=Self->Columns, colindex=0; col; col=col->Next, colindex++) {
      if (StrCompare(Args->Column, col->Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         sort.Column     = colindex;
         sort.Descending = Args->Descending;
         return(VIEW_SortColumnIndex(Self, &sort));
      }
   }

   return(ERR_Search);
}

/*****************************************************************************

###METHOD###
Name:  SortColumnIndex
Short: Sorts the view by column (index).

###DESCRIPTION###
<p>The SortColumnIndex() method will sort the view by the column index that
you indicate.  The sort will be ascending by default, which you can reverse
by setting the Descending parameter to TRUE.  The view will automatically
redraw itself to reflect the newly sorted content.</p>

<p>This method will fail if the referenced index is invalid.  Both hidden
and visible columns are taken into account when determining the column
that the index refers to.</p>

###INPUT###
LONG/Column: The index of the column that you want to sort by (starts from zero).
LONG/Descending: Set to TRUE if the sort should be descending.

###END###

*****************************************************************************/

static ERROR VIEW_SortColumnIndex(struct View *Self, struct mtSortColumnIndex *Args)
{
   WORD j;

   if ((!Args) OR (Args->Column < 0)) return(ERR_Args);

   /*** Extend the sort list so that we have a history of sort attempts ***/

   j = Self->Sort[0];
   if (j < 0) j = -j;
   if (j-1 != Args->Column) {
      for (j=ARRAYSIZE(Self->Sort)-1; j > 0; j--) Self->Sort[j] = Self->Sort[j-1];
   }

   if (Args->Descending IS TRUE) Self->Sort[0] = -(Args->Column + 1);
   else Self->Sort[0] = Args->Column + 1;

   /*** Do the sort based on the content of the Sort array ***/

   SortItems(Self);

   /*** Rearrange the view items and do a complete redraw ***/

   ArrangeItems(Self);
   ActionMsg(AC_Draw, Self->DrawableID, NULL);

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Sort
Short: Resorts XML data in the view when it has been manually altered.
###END###

*****************************************************************************/

static ERROR VIEW_Sort(struct View *Self, APTR Void)
{
   SortItems(Self);
   return(ERR_Okay);
}

/******************************************************************************

###ACTION###
Name:  UserClick
Short: User-click emulations can be passed through this action.
###END###

******************************************************************************/

static ERROR VIEW_UserClick(struct View *Self, struct acUserClick *Args)
{
   struct XMLTag **xmllist, *tag;
   struct ViewNode *node;
   struct Column *col;
   WORD i;
   BYTE active;
   LONG x, itemx, itemy;

   if (!Args) return(ERR_Args);

   Self->ClickX      = Args->XCoord;
   Self->ClickY      = Args->YCoord;
   Self->ClickIndex  = -1;
   Self->ActiveTag   = -1;
   tag = NULL;
   active = FALSE;

   if ((Self->Style IS VIEW_COLUMN) AND (Args->ButtonFlags & JD_LMB) AND
       (Args->YCoord >= Self->AreaY) AND (Args->YCoord < Self->AreaY + Self->ColumnHeight)) {
      /* Check if the click was over one of the column buttons.  This will
      ** affect sorting, or it can indicate a column resize.
      */

      x = Self->AreaX + Self->XPosition;
      for (col=Self->Columns, i=0; col; col=col->Next, i++) {
         if ((x+col->Width >= Self->ClickX-4) AND (x+col->Width < Self->ClickX+4)) {
            /*** The user has opted to resize the column ***/

            Self->ColumnResize = col;
            if (SetCursor(NULL, ACF_LMB, PTR_SPLIT_HORIZONTAL, NULL, Self->Head.UniqueID, NULL) IS ERR_Okay) {
               Self->PointerLocked = PTR_SPLIT_HORIZONTAL;
            }

            return(ERR_Okay);
         }
         else if ((Self->ClickX >= x) AND (Self->ClickX < x + col->Width)) {
            if (Self->Sort[0] IS i+1) ActionTags(MT_SortColumnIndex, Self, i, TRUE);
            else ActionTags(MT_SortColumnIndex, Self, i, FALSE);
            return(ERR_Okay);
         }
         x += col->Width;
      }
   }

   if (Self->Flags & VWF_NOSELECT) return(ERR_Okay);

   if ((GetField(Self->XML, FID_Tags, FT_POINTER, &xmllist) IS ERR_Okay) AND (xmllist)
       AND (Args->ButtonFlags & JD_LMB)) {

      for (tag=xmllist[0]; tag; tag=tag->Next) {
         node = tag->Private;
         itemx = Self->AreaX + node->XCoord + Self->XPosition;
         itemy = Self->AreaY + node->YCoord + Self->YPosition;
         if ((Args->XCoord >= itemx) AND (Args->XCoord < itemx + node->Width) AND
             (Args->YCoord >= itemy) AND (Args->YCoord < itemy + node->Height)) {

            Self->Buttons |= Args->ButtonFlags;

            if (node->Flags & NODE_SELECTED) {
               /* In multi-select mode, when the item is already selected
               ** we will deselect it.
               */

               Self->ActiveTag = tag->Index; /* The ActiveTag refers to the most recently selected OR -deselected- item */

               if (Args->ButtonFlags & JD_DBLCLICK) break;

               if (Self->Flags & VWF_MULTISELECT) {
                  node->Flags &= ~NODE_SELECTED;
                  DrawItem(Self, tag);
                  Self->Deselect = TRUE;
               }
            }
            else {
               active = SelectItem(Self, tag, TRUE);
               Self->Deselect = FALSE; /* We are not in deselect mode for click-dragging */
            }

            break;
         }
      }

      /*** Check the validity of the most current selected item ***/

      CheckSelectedItems(Self, xmllist[0]);
   }

   /* Double clicking will activate in item, if not already activated
   ** by SelectItem().
   */

   if ((Self->ActiveTag != -1) AND (Args->ButtonFlags & JD_LMB) AND (Args->ButtonFlags & JD_DBLCLICK) AND (active IS FALSE)) {
      acActivate(Self);
   }

   return(ERR_Okay);
}

/******************************************************************************

###ACTION###
Name:  UserClickRelease
Short: User-click emulations can be passed through this action.
###END###

******************************************************************************/

static ERROR VIEW_UserClickRelease(struct View *Self, struct acUserClickRelease *Args)
{
   if (!Args) return(ERR_Args);

   DPrintF("~ClickRelease()","[View]");

   Self->Buttons &= ~Args->ButtonFlags;
   Self->ClickIndex = -1;
   if (Args->ButtonFlags & JD_LMB) {
      Self->ColumnResize = NULL;

      /*** The cursor image must be checked following a click-release ***/

      CheckPointerCursor(Self, Args->XCoord, Args->YCoord);
   }

   StepBack();

   return(ERR_Okay);
}

/******************************************************************************

###ACTION###
Name:  UserMovement
Short: User movement emulations can be passed through this action.
###END###

******************************************************************************/

static ERROR VIEW_UserMovement(struct View *Self, struct acUserMovement *Args)
{
   struct XMLTag **xmllist, *tag, *scan, *checktag;
   struct ViewNode *node;
   struct Column *col;
   LONG cx, x, y, width, tagcount;
   WORD i, lastindex;
   BYTE highlighted, cursor;

   if (!Args) return(ERR_Args);

   if ((GetFields(Self->XML, FID_Tags|TPTR,      &xmllist,
                             FID_TagCount|TLONG, &tagcount,
                             TAGEND) != ERR_Okay) OR (!xmllist)) {
      return(ERR_Okay);
   }

   /*** Check the ClickIndex field to make sure it's not lying outside of the XML tag array ***/

   if (Self->ClickIndex > tagcount) Self->ClickIndex = -1;

   cursor = FALSE;

   CheckPointerCursor(Self, Args->XCoord, Args->YCoord);

   /*** Handle column resizing ***/

   x = Args->XCoord;
   y = Args->YCoord;

   if ((Self->Style IS VIEW_COLUMN) AND (Self->ColumnResize)) {
      /* Calculate the horizontal position of the
      ** column being resized.
      */

      cx = 0;
      for (col=Self->Columns; col != Self->ColumnResize; col=col->Next) cx += col->Width;

      if (col) {
         width = col->Width;
         col->Width = x - Self->XPosition - cx;
         if (col->Width < MIN_COLWIDTH) col->Width = MIN_COLWIDTH;
         if (col->Width != width) {
            /* Recalculate the width of the scrollable page and reset the widths of
            ** each item to match the page width.
            */

            for (col=Self->Columns, Self->PageWidth=0; col; col=col->Next) Self->PageWidth += col->Width;

            ArrangeItems(Self); /* Rearrange items to match the new column/page width */

            /*** Recalculate the horizontal scrollbar, then issue a redraw ***/

            CalculateHScroll(Self);

            ActionMsg(AC_Draw, Self->DrawableID, NULL);
         }
         return(ERR_Okay);
      }
      else Self->ColumnResize = NULL;
   }

   /*** Adjust for scrolling ***/

   x -= Self->XPosition;
   y -= Self->YPosition;

   /* If the mouse coordinates fall outside of the view area, restrict them.  This
   ** is useful when the user drags the pointer while the mouse button is held down.
   */

   if (x < Self->AreaX) x = Self->AreaX;
   if (y < Self->AreaY + Self->ColumnHeight) y = Self->AreaY + Self->ColumnHeight + (Self->LineHeight / 2);

   if (x > Self->PageWidth) x = Self->PageWidth;
   if (y > Self->AreaY + Self->PageHeight) y = Self->AreaY + Self->PageHeight - (Self->LineHeight / 2);

   highlighted = FALSE;
   checktag = NULL;
   for (tag=xmllist[0]; tag; tag=tag->Next) {
      node = tag->Private;

      if ((x >= Self->AreaX + node->XCoord) AND (x < Self->AreaX + node->XCoord + node->Width) AND
          (y >= Self->AreaY + node->YCoord) AND (y < Self->AreaY + node->YCoord + node->Height)) {

         if (Self->Buttons & JD_LMB) {
            checktag = tag;

            if (Self->ClickIndex IS -1) Self->ClickIndex = tag->Index;

            if ((Self->Flags & VWF_SENSITIVE) OR (!(Self->Flags & VWF_MULTISELECT))) {

               //DPrintF("UserMovement:","[View] Selecting tag %d", tag->Index);

               if (!(node->Flags & NODE_SELECTED)) {
                  /* Scan for any existing selections and
                  ** turn them off.
                  */

                  for (scan=xmllist[0]; scan; scan=scan->Next) {
                     node = scan->Private;
                     if (node->Flags & NODE_SELECTED) {
                        node->Flags &= ~NODE_SELECTED;
                        DrawItem(Self, scan);
                     }
                  }
                  node = tag->Private;

                  node->Flags |= NODE_SELECTED;
                  DrawItem(Self, tag);

                  /*** Change the currently active tag ***/

                  SetFields(Self, FID_SelectedTag|TLONG, tag->Index, TAGEND);

                  if (!(Self->Flags & VWF_MULTISELECT)) {
                     Self->HighlightTag = tag->Index;
                     Self->ActiveTag = tag->Index;
                  }

                  /*** Activate if sensitive mode is enabled ***/

                  if (Self->Flags & VWF_SENSITIVE) {
                     /*** Activate so long as the 'insensitive' attribute has not been set against the tag ***/

                     for (i=0; i < tag->TotalAttributes; i++) {
                        if (StrCompare(tag->Attributes[i].Name, "insensitive", 0, STR_MATCHLENGTH) IS ERR_Okay)  break;
                     }

                     if (i >= tag->TotalAttributes) acActivate(Self);
                  }
               }
            }
            else {
               /* Select or deselect everything between the item at which the LMB
               ** was held and the item where we are at now.
               */

               //DPrintF("UserMovement:","[View] Single-select for tag %d", tag->Index);

               i = Self->ClickIndex; /* ClickIndex is the item at which the LMB was held */
               lastindex = tag->Index; /* LastIndex is the item at which we are now */

               for (tag=xmllist[Self->ClickIndex]; tag; ) {
                  node = tag->Private;

                  if (node->Flags & NODE_SELECTED) {
                     /*** The node is currently selected ***/
                     if (Self->Deselect IS TRUE) {
                        node->Flags &= ~NODE_SELECTED;
                        DrawItem(Self, tag);
                     }
                  }
                  else {
                     /*** The node is not yet selected ***/
                     if (Self->Deselect IS FALSE) {
                        node->Flags |= NODE_SELECTED;
                        DrawItem(Self, tag);
                     }
                  }

                  if (tag->Index IS lastindex) break;

                  if (tag->Index < lastindex) tag = tag->Next; /* Scan forwards */
                  else tag = tag->Prev; /* Scan backwards */
               }
            }
         }
         else if ((Args->XCoord - Self->XPosition >= node->XCoord) AND (Args->XCoord - Self->XPosition < node->XCoord + node->Width) AND
                  (Args->YCoord - Self->YPosition >= node->YCoord) AND (Args->YCoord - Self->YPosition < node->YCoord + node->Height) AND
                  (Args->OverObject IS Self->DrawableID) AND (!Self->Buttons)) {

            /*** Highlight the underlying item (do not select it) ***/

            if (Self->HighlightTag != tag->Index) {
               LONG oldtag;
               if ((oldtag = Self->HighlightTag) != -1) {
                  Self->HighlightTag = -1;
                  DrawItem(Self, xmllist[oldtag]);
               }

               Self->HighlightTag = tag->Index;
               DrawItem(Self, tag);
            }
            highlighted = TRUE;
         }
         break;
      }
   }

   /*** Unhighlight the most recently highlighted item ***/

   if ((highlighted IS FALSE) AND (Self->HighlightTag != -1)) {
      tag = xmllist[Self->HighlightTag];
      Self->HighlightTag = -1;
      DrawItem(Self, tag);
   }

   CheckSelectedItems(Self, xmllist[0]);

   if (checktag) CheckItemVisible(Self, checktag);

   StepBack();

   return(ERR_Okay);
}

/******************************************************************************
** Internal: CheckSelectedItems()
**
** This function checks that the SelectedTag refers to a valid, selected item.
** If it doesn't, the SelectedTag field is recalculated.
*/

static void CheckSelectedItems(struct View *Self, struct XMLTag *Tags)
{
   struct XMLTag *scan, **xmllist;
   struct ViewNode *node;

   if (Self->SelectedTag IS -1) return;

   if ((GetField(Self->XML, FID_Tags, FT_POINTER, &xmllist) != ERR_Okay) OR (!xmllist)) return;

   node = xmllist[Self->SelectedTag]->Private;
   if (node->Flags & NODE_SELECTED) return;

   for (scan=Tags; scan; scan=scan->Next) {
      node = scan->Private;
      if (node->Flags & NODE_SELECTED) {
         SetFields(Self, FID_SelectedTag|TLONG, scan->Index, TAGEND);
         return;
      }
   }

   Self->SelectedTag = -1;
}

/*****************************************************************************

###FIELD###
Name:   BorderOffset
Short:  Sets the X, Y, XOffset and YOffset to a single value.
Type:   LONG
Status: Set

###DESCRIPTION###
<p>This field is provided for the convenience of setting the X, Y, XOffset
and YOffset fields in one hit.  The value that you specify will be written
to all of the aforementioned fields as a fixed pixel-width value.</p>

###END###

*****************************************************************************/

static ERROR SET_BorderOffset(struct View *Self, LONG *Value)
{
   SetFields(Self, FID_XCoord|TLONG,  *Value,
                   FID_YCoord|TLONG,  *Value,
                   FID_XOffset|TLONG, *Value,
                   FID_YOffset|TLONG, *Value,
                   TAGEND);
   return(ERR_Okay);
}

static ERROR SET_ColBackground(struct View *Self, STRING Colour)
{
   StrReadColour(Colour, &Self->ColBackground.Red, &Self->ColBackground.Green, &Self->ColBackground.Blue, &Self->ColBackground.Alpha);
   return(ERR_Okay);
}

static ERROR SET_ColHighlight(struct View *Self, STRING Colour)
{
   StrReadColour(Colour, &Self->ColHighlight.Red, &Self->ColHighlight.Green, &Self->ColHighlight.Blue, &Self->ColHighlight.Alpha);
   return(ERR_Okay);
}

static ERROR SET_ColShadow(struct View *Self, STRING Colour)
{
   StrReadColour(Colour, &Self->ColShadow.Red, &Self->ColShadow.Green, &Self->ColShadow.Blue, &Self->ColShadow.Alpha);
   return(ERR_Okay);
}

static ERROR SET_Hairline(struct View *Self, STRING Colour)
{
   StrReadColour(Colour, &Self->HairlineRGB.Red, &Self->HairlineRGB.Green, &Self->HairlineRGB.Blue, &Self->HairlineRGB.Alpha);
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Columns
Short:  Active columns for 'column mode' may be set via this field.
Type:   STRING
Status: Get/Set

###DESCRIPTION###
<p>When a view object is in column or tree mode (as defined in the <a href="#Style">Style</a> field), you
need to tell the view object what XML tags should be translated into column areas.
Failure to register any column names means that only the default column will
be displayed.  Examine the following XML structure:</p>

<pre>
  &lt;file&gt;
    readme.txt
    &lt;size&gt;19431&lt;/size&gt;
    &lt;date&gt;20030503 11:53:19&lt;/date&gt;
  &lt;/file&gt;
</pre>

<p>The size and date tags will need to be registered as columns in order for
the view object to display them.  To show them in the order specified, we
can pass the string "size(text:'File Size', len:140, type:bytesize);
date(text:Date, len:180, type:date)" to the Columns field.
The optional information that we've specified in the brackets tells the view
object the column name, the default pixel width of the column and the display
type.</p>

<p>Currently recognised data types are 'bytesize', 'numeric' and 'date'.  Use
the bytesize type if the value is a number that can be represented in bytes,
kilobytes, megabytes or gigabytes.  The date type can be used if the values
are stored in the date format "YYYYMMDD HH:MM:SS".  The numeric type can be
used for any type of number (integer or floating point).  If the type is not
specified, the view object will assume that the column is a standard string
or variant type.</p>

<p>If you would like to change the arrangement of the default column so that
it appears at a different position, you may refer to it using a column name of
'default'.</p>

###END###

*****************************************************************************/

static ERROR SET_Columns(struct View *Self, STRING Value)
{
   struct Column col, *newcol, *column, *prevcol, *scan;
   STRING str;
   UBYTE arg[20], buffer[120];
   WORD i, index, colindex;
   BYTE reset;

   DPrintF("~SetField:","[View.Columns] %s", Value);

   /*** Mark all existing columns for deletion ***/

   for (column=Self->Columns; column; column=column->Next) column->Flags |= CF_DELETE;

   index = 0;
   str = Value;
   while (*str) {
      ClearMemory(&col, sizeof(col));

      while ((*str) AND (*str <= 0x20)) str++; /* Skip whitespace */

      /*** Extract the name ***/

      for (i=0; (*str) AND (*str != '(') AND (*str != ';') AND (i < sizeof(col.Name)-1); i++) {
         col.Name[i] = *str;
         str++;
      }
      col.Name[i]  = 0;

      if (i < 1) break;

      reset = FALSE;
      if (*str IS '(') {
         str++;

         while ((*str) AND (*str != ')')) {
            while ((*str) AND (*str <= 0x20)) str++; /* Skip whitespace */

            for (i=0; (*str) AND (i < sizeof(arg)-1) AND (*str != ':') AND (*str != ',') AND (*str != ')'); i++) arg[i] = *str++;
            arg[i] = 0;

            buffer[0] = 0;
            if (*str IS ':') {
               *str++;

               if (*str IS '"') {
                  *str++;
                  for (i=0; (*str) AND (*str != '"') AND (i < sizeof(buffer)-1); i++) {
                     buffer[i] = *str;
                     str++;
                  }
                  buffer[i] = 0;

                  if (*str IS '"') str++;
               }
               else if (*str IS '\'') {
                  *str++;
                  for (i=0; (*str) AND (*str != '\'') AND (i < sizeof(buffer)-1); i++) {
                     buffer[i] = *str;
                     str++;
                  }
                  buffer[i] = 0;

                  if (*str IS '\'') str++;
               }
               else {
                  for (i=0; (*str) AND (*str != ')') AND (*str != ',') AND (i < sizeof(buffer)-1); i++) {
                     buffer[i] = *str;
                     str++;
                  }
                  buffer[i] = 0;
               }
            }

            while ((*str) AND (*str != ',') AND (*str != ';') AND (*str != ')')) str++;
            if ((*str IS ',') OR (*str IS ';')) str++;

            if (arg[0]) {
               if (StrCompare("text", arg, 0, STR_MATCHLENGTH) IS ERR_Okay) {
                  StrCopy(buffer, col.Text, sizeof(col.Text));
               }
               else if (StrCompare("len", arg, 0, STR_MATCHLENGTH) IS ERR_Okay) {
                  col.Width = StrToInt(buffer);
               }
               else if (StrCompare("reset", arg, 0, STR_MATCHLENGTH) IS ERR_Okay) {
                  reset = TRUE;
               }
               else if (StrCompare("type", arg, 0, STR_MATCHLENGTH) IS ERR_Okay) {
                  if (StrCompare("numeric", buffer, 0, NULL) IS ERR_Okay) col.Type = CT_NUMERIC;
                  else if (StrCompare("number", buffer, 0, NULL) IS ERR_Okay) col.Type = CT_NUMERIC;
                  else if (StrCompare("integer", buffer, 0, NULL) IS ERR_Okay) col.Type = CT_NUMERIC;
                  else if (StrCompare("date", buffer, 0, NULL) IS ERR_Okay) col.Type = CT_DATE;
                  else if (StrCompare("bytesize", buffer, 0, NULL) IS ERR_Okay) col.Type = CT_BYTESIZE;
               }
               else if ((StrCompare("colour", arg, 0, STR_MATCHLENGTH) IS ERR_Okay) OR
                        (StrCompare("col", arg, 0, STR_MATCHLENGTH) IS ERR_Okay)) {
                  col.Flags |= CF_COLOUR;
                  StrReadColour(buffer, &col.Colour.Red, &col.Colour.Green, &col.Colour.Blue, &col.Colour.Alpha);
               }
               else if (StrCompare("showicons", arg, 0, STR_MATCHLENGTH) IS ERR_Okay) {
                  col.Flags |= CF_SHOWICONS;
               }
               else if (StrCompare("rightalign", arg, 0, STR_MATCHLENGTH) IS ERR_Okay) {
                  col.Flags |= CF_RIGHTALIGN;
               }
               else DPrintF("@View:","Unsupported column argument '%s'", arg);
            }
         }
      }

      /*** Scan up to the next entry ***/

      while ((*str) AND (*str <= 0x20)) str++; /* Skip whitespace */
      while ((*str) AND (*str != ';')) str++; /* Look for a semi-colon */
      if (*str IS ';') str++;
      while ((*str) AND (*str <= 0x20)) str++; /* Skip whitespace */

      if (!col.Text[0]) StrCopy(col.Name, col.Text, sizeof(col.Text));
      if (!col.Width) col.Width = 100;

      if (col.Width < MIN_COLWIDTH) col.Width = MIN_COLWIDTH;

      /* Check if the column already exists in our array.  If it does,
      ** update its details.  Otherwise we will create a new
      ** column and add it to the column table.
      */

      prevcol = NULL;
      for (column=Self->Columns, colindex=0; column; column=column->Next, colindex++) {
         if (StrCompare(column->Name, col.Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
            DPrintF("SetField:","Updating column '%s'", col.Name);

            StrCopy(col.Text, column->Text, sizeof(column->Text));
            if ((reset) OR (!(Self->Head.ObjectFlags & NF_INITIALISED))) {
               column->Width = col.Width;  /* Allow the user to retain his defined width unless the reset option has been forced */
            }
            if (col.Type)  column->Type  = col.Type;
            if (col.Sort)  column->Sort  = col.Sort;
            if (col.Flags) column->Flags = col.Flags;
            if (col.Flags & CF_COLOUR) column->Colour = col.Colour;
            column->Flags &= ~CF_DELETE;

            if (colindex != index) {
               /* The order of the columns needs to be altered.  First of all,
               ** we will patch the list as if the column is no longer available.
               */

               if (prevcol) prevcol->Next = column->Next;
               else Self->Columns = column->Next;

               /*** Now re-insert the column at the required position ***/

               prevcol = NULL;
               for (colindex=0, scan=Self->Columns; (scan) AND (colindex < index); scan=scan->Next, colindex++) prevcol = scan;
               column->Next = scan;
               if (prevcol) prevcol->Next = column;
               else Self->Columns = column;
            }

            break;
         }
         prevcol = column;
      }

      /*** Add the column if an existing match wasn't found ***/

      if (!column) {
         DPrintF("SetField:","Adding new column '%s', type %d", col.Name, col.Type);
         if (AllocMemory(sizeof(col), MEM_DATA|MEM_NOCLEAR, &newcol, NULL) IS ERR_Okay) {
            CopyMemory(&col, newcol, sizeof(col));
            if (!Self->Columns) Self->Columns = newcol;
            else {
               for (column=Self->Columns; column->Next; column=column->Next);
               column->Next = newcol;
            }
         }
      }

      index++;
   }

   /*** Remove columns marked for deletion ***/

   scan = NULL;
   prevcol = NULL;
   column = Self->Columns;
   while (column) {
      scan = column;
      column = column->Next;
      if (scan->Flags & CF_DELETE) {
         DPrintF("SetField:","Deleting column '%s'", scan->Name);
         if (scan IS Self->Columns) Self->Columns = scan->Next;
         else if (prevcol) prevcol->Next = scan->Next;
         FreeMemory(scan);
      }
      else prevcol = scan;
   }

   /*** Clear sort settings ***/

   ClearMemory(Self->Sort, sizeof(Self->Sort));

   /*** Save column header settings ***/

   if (Self->ColumnString) FreeMemory(Self->ColumnString);
   Self->ColumnString = StrClone(Value);

   /*** Redraw the view if we are in column mode ***/

   if (Self->Style IS VIEW_COLUMN) {
      ArrangeItems(Self);
      ActionMsg(AC_Draw, Self->DrawableID, NULL);
   }

   StepBack();
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   DateFormat
Short:  Sets the format to use when displaying date types.
Type:   STRING
Status: Get/Set

###DESCRIPTION###
<p>Use the DateFormat field to define how dates should be displayed when they
are used in columns.  For information on how to set valid date formats,
refer to the StrFormatDate() function in the Strings module.</p>

<p>If you do not set this field then the user's preferred date format will
be used.</p>

###END###

*****************************************************************************/

static ERROR GET_DateFormat(struct View *Self, STRING *Value)
{
   *Value = Self->DateFormat;
   return(ERR_Okay);
}

static ERROR SET_DateFormat(struct View *Self, STRING Value)
{
   if ((Value) AND (Value[0])) StrCopy(Value, Self->DateFormat, sizeof(Self->DateFormat));
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Height
Short:  Defines the height of a view.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>A view can be given a fixed or relative height by setting this field
to the desired value.  To set a relative height, use the FD_PERCENT flag when
setting the field.</p>

###END###

*****************************************************************************/

static ERROR GET_Height(struct View *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   LONG containerheight, ycoord;
   DOUBLE value;

   if (AccessObject(Self->DrawableID, 4000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_Height, FT_LONG, &containerheight);
      ReleaseObject(drawable);

      if (containerheight <= 0) return(ObjectError(Self, ERH_GetField, ERR_GetField));

      if (Self->Dimensions & DMF_FIXED_HEIGHT) value = Self->Height;
      else if (Self->Dimensions & DMF_RELATIVE_HEIGHT) {
         value = Self->Height * containerheight / 100;
      }
      else if ((Self->Dimensions & (DMF_FIXED_Y|DMF_RELATIVE_Y)) AND
               (Self->Dimensions & (DMF_FIXED_YOFFSET|DMF_RELATIVE_YOFFSET))) {
         if (Self->Dimensions & DMF_FIXED_Y) ycoord = Self->YCoord;
         else ycoord = (containerheight * Self->YCoord) / 100;
         if (Self->Dimensions & DMF_FIXED_YOFFSET) value = containerheight - ycoord - Self->YOffset;
         else value = containerheight - ycoord - ((containerheight * Self->YOffset) / 100);
      }
      else return(ObjectError(Self, ERH_GetHeight, ERR_InvalidDimension));

      if (Value->Type & FD_PERCENTAGE) value = (value * 100) / containerheight;
   }
   else return(ObjectError(Self, ERH_GetHeight, ERR_AccessObject));

   if (Value->Type & FD_DOUBLE) Value->Double = value;
   else if (Value->Type & FD_LARGE) Value->Large = value;
   else return(ObjectError(Self, ERH_GetHeight, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

static ERROR SET_Height(struct View *Self, struct Variable *Value)
{
   if (Value->Type & FD_DOUBLE) {
      if (Value->Double <= 0) {
         DPrintF("@SetField:","[View] A height of %.2f is illegal.", Value->Double);
         return(ERR_OutOfRange);
      }
      Self->Height = Value->Double;
   }
   else if (Value->Type & FD_LARGE) {
      if (Value->Large <= 0) {
         DPrintF("@SetField:","[View] A height of %d is illegal.", Value->Large);
         return(ERR_OutOfRange);
      }
      Self->Height = Value->Large;
   }
   else return(ObjectError(Self, ERH_SetHeight, ERR_FieldTypeMismatch));

   if (Value->Type & FD_PERCENTAGE) Self->Dimensions = (Self->Dimensions & ~DMF_FIXED_HEIGHT) | DMF_RELATIVE_HEIGHT;
   else Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_HEIGHT) | DMF_FIXED_HEIGHT;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   HighlightColour
Short:  Defines the colour when an item is selected via mouse-over or user focus.
Type:   STRING
Status: Set

###DESCRIPTION###
<p>The HighlightColour defines the colour that is used when the user rolls
the mouse pointer over an item, or uses the cursor keys to change the item
focus.  The colour must be in hexadecimal or CSV format - for example to
create a pure red colour, a setting of &quot;#ff0000&quot; or
&quot;255,0,0&quot; would be valid.</p>

###SEE ALSO###
Field: SelectColour

###END###

*****************************************************************************/

static ERROR SET_HighlightColour(struct View *Self, STRING Colour)
{
   StrReadColour(Colour, &Self->HighlightRGB.Red, &Self->HighlightRGB.Green, &Self->HighlightRGB.Blue, &Self->HighlightRGB.Alpha);
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   HScroll
Short:  If scrolling is required, use this field to refer to a horizontal scroll bar.
Type:   OBJECTID
Status: Read/Set

###DESCRIPTION###
<p>If you want to attach a horizontal scrollbar to a view object, set
this field to an object belonging to the Scroll class.  So long as the Scroll
object is set up to provide full scrollbar functionality, the user will be
able to scroll the text display along the horizontal axis.</p>

###SEE ALSO###
Field: VScroll
Class: Scroll

###END###

*****************************************************************************/

static ERROR SET_HScroll(struct View *Self, OBJECTID *ScrollID)
{
   OBJECTPTR scroll;

   if (GetClassID(*ScrollID) IS ID_SCROLL) {
      if (AccessObject(*ScrollID, 3000, &scroll) IS ERR_Okay) {
         SetField(scroll, FID_Object, FT_LONG, &Self->Head.UniqueID);
         Self->HScroll = *ScrollID;
         Self->XPosition = 0;
         CalculateHScroll(Self);
         ReleaseObject(scroll);
         return(ERR_Okay);
      }
      else return(ObjectError(Self, ERH_SetField, ERR_ExclusiveDenied));
   }
   else {
      DPrintF("@SetField:","[Text] Attempt to set the HScroll field with an invalid object.");
      return(ERR_Failed);
   }
}

/*****************************************************************************

###FIELD###
Name:   ItemColour
Short:  The default font colour to use for view items.
Type:   STRING
Status: Set

###DESCRIPTION###
<p>The ItemColour defines the font colour that is applied to new items.  The
default is black.  Items may override the default by using the colour attribute
in their XML definition.</p>

<p>The colour string must be in hexadecimal or CSV format - for example to
create a pure red colour, a setting of &quot;#ff0000&quot; or
&quot;255,0,0&quot; would be valid.</p>

###SEE ALSO###
Field: HighlightColour, SelectColour

###END###

*****************************************************************************/

static ERROR SET_ItemColour(struct View *Self, STRING Colour)
{
   StrReadColour(Colour, &Self->ItemRGB.Red, &Self->ItemRGB.Green, &Self->ItemRGB.Blue, &Self->ItemRGB.Alpha);
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   SelectColour
Short:  Defines the colour when an item is selected via mouse-over or user focus.
Type:   STRING
Status: Set

###DESCRIPTION###
<p>The SelectColour defines the colour that is used when the user rolls
the mouse pointer over an item, or uses the cursor keys to change the item
focus.  The colour must be in hexadecimal or CSV format - for example to
create a pure red colour, a setting of &quot;#ff0000&quot; or
&quot;255,0,0&quot; would be valid.</p>

###SEE ALSO###
Field: SelectColour

###END###

*****************************************************************************/

static ERROR SET_SelectColour(struct View *Self, STRING Colour)
{
   StrReadColour(Colour, &Self->SelectRGB.Red, &Self->SelectRGB.Green, &Self->SelectRGB.Blue, &Self->SelectRGB.Alpha);
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   SelectedTags
Short:  Returns an array of tag indexes for all currently selected items.
Type:   LONG *
Status: Get

###DESCRIPTION###
<p>A complete list of currently selected items can be retrieved by reading
the SelectedTags field.  The data is returned as an array of 32-bit tag indexes,
terminated with a value of -1.  An error code will be returned if there are no
currently selected items.</p>

<p>The array remains valid up until the next time that you read the SelectedTags
field.</p>

###END###

*****************************************************************************/

static ERROR GET_SelectedTags(struct View *Self, LONG **Array)
{
   struct XMLTag **xmllist, *tag;
   struct ViewNode *node;
   OBJECTPTR context;
   ERROR error;
   WORD i, count;

   *Array = NULL;

   if (Self->SelectedTags) {
      FreeMemory(Self->SelectedTags);
      Self->SelectedTags = NULL;
   }

   if ((GetField(Self->XML, FID_Tags, FT_POINTER, &xmllist) != ERR_Okay) OR (!xmllist)) return(ERR_NoData);

   /*** Count the total number of selected items ***/

   count = 0;
   for (tag=xmllist[0]; tag; tag = tag->Next) {
      node = tag->Private;
      if (node->Flags & NODE_SELECTED) count++;
   }

   if (count < 1) return(ERR_NoData);

   /*** Allocate the array and fill it ***/

   context = SetContext(Self);
   error = AllocMemory(sizeof(LONG) * (count + 1), MEM_DATA|MEM_NOCLEAR, &Self->SelectedTags, NULL);
   SetContext(context);

   if (error IS ERR_Okay) {
      i = 0;
      for (tag=xmllist[0]; tag; tag = tag->Next) {
         node = tag->Private;
         if (node->Flags & NODE_SELECTED) {
            Self->SelectedTags[i] = tag->Index;
            i++;
         }
      }
      Self->SelectedTags[i] = -1;

      *Array = Self->SelectedTags;
      return(ERR_Okay);
   }
   else return(ObjectError(Self, ERH_GetField, ERR_AllocMemory));
}

/*****************************************************************************

###FIELD###
Name:   Selection
Short:  Indicates the currently selected file or directory.
Type:   STRING
Status: Get/Set

###DESCRIPTION###
<p>The Selection field provides a method for retrieving the content of the
currently selected item.  If no selection is active then this field will
return an error code and a NULL pointer.</p>

<p>You can also manually select an item by writing a valid item string to
this field.  The string must refer to the name of an item in the default
column of the view.</p>

###END###

*****************************************************************************/

static ERROR GET_Selection(struct View *Self, STRING *Value)
{
   if (Self->SelectedTag != -1) {
      if (ActionTags(MT_GetXMLContent, Self->XML, Self->SelectedTag, &Self->Selection, sizeof(Self->Selection)) IS ERR_Okay) {
         *Value = Self->Selection;
         return(ERR_Okay);
      }
   }

   *Value = NULL;
   return(ERR_NoData);
}

static ERROR SET_Selection(struct View *Self, STRING Value)
{
   struct XMLTag *tag, **taglist;
   UBYTE buffer[300];

   DPrintF("~SetField:","[View.Selection] = %s", Value);

   if ((GetFields(Self->XML, FID_Tags|TPTR, &taglist, TAGEND) IS ERR_Okay) AND (taglist)) {
      for (tag=*taglist; tag; tag=tag->Next) {
         if (ActionTags(MT_GetXMLContent, Self->XML, tag->Index, buffer, sizeof(buffer)) IS ERR_Okay) {
            if (StrCompare(Value, buffer, 0, STR_MATCHLENGTH) IS ERR_Okay) {
               SelectItem(Self, tag, TRUE);
               StepBack();
               return(ERR_Okay);
            }
         }
      }
   }

   DPrintF("@SetField:","[View.Selection] Unable to find item \"%s\"", Value);
   StepBack();
   return(ERR_Search);
}

/*****************************************************************************

###FIELD###
Name:   SelectionIndex
Short:  Indicates the index number of the currently selected file or directory.
Type:   STRING
Status: Get/Set

###DESCRIPTION###
<p>The SelectionIndex field provides a method for retrieving the index of the
currently selected item.  If no selection is active then this field will
return a value of -1.</p>

<p>You can also manually select an item by writing a valid index number to
this field.  Changing the index will cause the item to become highlighting
and the view's children will be activated automatically to inform them of
the change.  An error will be returned if the index number is outside of the available
range.  An index of -1 will deselect any currently selected items.</p>

###END###

*****************************************************************************/

static ERROR GET_SelectionIndex(struct View *Self, LONG *Value)
{
   struct XMLTag **xmllist, *tag;
   LONG index;

   if (Self->SelectedTag != -1) {
      if ((GetField(Self->XML, FID_Tags, FT_POINTER, &xmllist) IS ERR_Okay) AND (xmllist)) {
         index = 0;
         for (tag=xmllist[0]; (tag) AND (tag->Index != Self->SelectedTag); tag=tag->Next) index++;
         *Value = index;
         return(ERR_Okay);
      }
   }

   *Value = -1;
   return(ERR_Okay);
}

static ERROR SET_SelectionIndex(struct View *Self, LONG *Value)
{
   struct XMLTag *tag, **taglist;
   LONG index, count;

   if (*Value IS -1) {
      /*** Deselect all items ***/
      DPrintF("SetField:","[View.SelectionIndex] = %d (deselect-all)", *Value);
      SelectItem(Self, NULL, FALSE);
      return(ERR_Okay);
   }
   else if (Self->Head.ObjectFlags & NF_INITIALISED) {
      DPrintF("~SetField:","[View.SelectionIndex] = %d", *Value);
      if ((GetFields(Self->XML, FID_Tags|TPTR, &taglist, TAGEND) IS ERR_Okay) AND (taglist)) {
         index = *Value;
         count = 0;
         for (tag=*taglist; tag; tag=tag->Next) {
            if (!index) {
               SelectItem(Self, tag, TRUE);
               acActivate(Self);
               StepBack();
               return(ERR_Okay);
            }
            index--;
            count++;
         }

         DPrintF("@SetField:","[View.SelectionIndex] Index %d out of range (max %d).", *Value, count);
         StepBack();
         return(ERR_OutOfRange);
      }
      else {
         DPrintF("@SetField:","[View.SelectionIndex] No data in the view.");
         StepBack();
         return(ERR_NoData);
      }
   }
   else {
      Self->SelectionIndex = *Value; /* SelectionIndex field is provided purely for assisting the init process */
      return(ERR_Okay);
   }
}

/*****************************************************************************

###FIELD###
Name:   TotalSelected
Short:  The total number of currently selected items.
Type:   LONG
Status: Get

###DESCRIPTION###
<p>This readable field will tell you the total number of items that are
currently selected by the user.</p>

###END###

*****************************************************************************/

static ERROR GET_TotalSelected(struct View *Self, LONG *Value)
{
   struct XMLTag **xmllist, *tag;
   struct ViewNode *node;
   WORD count;

   *Value = NULL;

   if ((GetField(Self->XML, FID_Tags, FT_POINTER, &xmllist) != ERR_Okay) OR (!xmllist)) return(ERR_NoData);

   count = 0;
   for (tag=xmllist[0]; tag; tag = tag->Next) {
      node = tag->Private;
      if (node->Flags & NODE_SELECTED) count++;
   }

   *Value = count;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Style
Short:  The style of view that is displayed to the user is configured here.
Type:   LONG
Status: Read/Set

###DESCRIPTION###
<p>The style of the view display can be configured by writing this field.
You should set this field on initialisation, and may change it at any time
if the user wishes to switch to a different style.  Available view styles
are as follows:</p>

<table border="1" bordercolordark="#000000" cellspacing="0" cellpadding="2" bordercolorlight="#000000" width="80%" align="center">
<tr><td class="tableheader" bgcolor="#eeeeee">Flag</td><td class="tableheader" bgcolor="#eeeeee">Description</td></tr>
<tr><td width="10%">COLUMN</td><td>Column mode is used when items carry more than one element of information.  For instance, a file item may carry the file name, date and size, each of which can have its own column.  The format of the column display is refined in the Columns field.</td></tr>
<tr><td width="10%">ICON</td><td>Icon mode displays each item as a large icon with the default item text written underneath.</td></tr>
<tr><td width="10%">LIST</td><td>List mode displays items using small icons and writes the item text on the right.  The arrangement of the items is from top to bottom, left to right.</td></tr>
<tr><td width="10%">LONGLIST</td><td>This type is the same as the LIST view but the items will not wrap back to the top of the view when the bottom edge of the drawable is encountered.</td></tr>
<tr><td width="10%">TREE</td><td>Tree mode can be used if the items are arranged in a hierarchy, where items can be embedded within items.  The tree view provides small buttons that the user can click to expand the tree and view hidden items.</td></tr>
</table>

###END###

*****************************************************************************/

static ERROR SET_Style(struct View *Self, LONG *Value)
{
   if (Self->Style != *Value) {
      Self->Style = *Value;

      if (Self->Head.ObjectFlags & NF_INITIALISED) {
         DPrintF("~Set.Style:","[View] The view style has changed.");
         Self->XPosition = 0;
         Self->YPosition = 0;
         ArrangeItems(Self);
         DelayMsg(AC_Draw, Self->DrawableID, NULL);
         StepBack();
      }
   }

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   VScroll
Short:  If scrolling is required, use this field to refer to a vertical scroll bar.
Type:   OBJECTID
Status: Read/Set

###DESCRIPTION###
<p>If you want to attach a vertical scrollbar to a view object, set this
field to an object belonging to the Scroll class.  So long as the Scroll
object is configured to provide full scrollbar functionality, the user will be
able to scroll the text display along the vertical axis.</p>

###SEE ALSO###
Fields: HScroll
Class: Scroll

###END###

*****************************************************************************/

static ERROR SET_VScroll(struct View *Self, OBJECTID *ScrollID)
{
   OBJECTPTR scroll;
   OBJECTID scrollid;

   scrollid = *ScrollID;
   if (GetClassID(*ScrollID) != ID_SCROLL) {
      DPrintF("@SetField:","[View] Attempt to set the VScroll field with an invalid object.");
      return(ERR_Failed);
   }

   if (AccessObject(*ScrollID, 3000, &scroll) IS ERR_Okay) {
      SetField(scroll, FID_Object, FT_LONG, &Self->Head.UniqueID);
      Self->VScroll = *ScrollID;
      Self->YPosition = 0;
      CalculateVScroll(Self);
      ReleaseObject(scroll);
      return(ERR_Okay);
   }
   else return(ObjectError(Self, ERH_SetField, ERR_ExclusiveDenied));
}

/*****************************************************************************

###FIELD###
Name:   Width
Short:  Defines the width of a view.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>A view can be given a fixed or relative width by setting this field
to the desired value.  To set a relative width, use the FD_PERCENT flag when
setting the field.</p>

###END###

*****************************************************************************/

static ERROR GET_Width(struct View *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   LONG containerwidth;
   DOUBLE xcoord, value;

   if (AccessObject(Self->DrawableID, 4000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_Width, FT_LONG, &containerwidth);
      ReleaseObject(drawable);

      if (containerwidth <= 0) return(ObjectError(Self, ERH_GetWidth, ERR_GetField));

      if (Self->Dimensions & DMF_FIXED_WIDTH) value = Self->Width;
      else if (Self->Dimensions & DMF_RELATIVE_WIDTH) {
         value = Self->Width * containerwidth / 100;
      }
      else if ((Self->Dimensions & (DMF_FIXED_X|DMF_RELATIVE_X)) AND
               (Self->Dimensions & (DMF_FIXED_XOFFSET|DMF_RELATIVE_XOFFSET))) {
         if (Self->Dimensions & DMF_FIXED_X) xcoord = Self->XCoord;
         else xcoord = containerwidth * Self->XCoord / 100;
         if (Self->Dimensions & DMF_FIXED_XOFFSET) value = containerwidth - xcoord - Self->XOffset;
         else value = containerwidth - xcoord - (containerwidth * Self->XOffset / 100);
      }
      else return(ObjectError(Self, ERH_GetWidth, ERR_InvalidDimension));

      if (Value->Type & FD_PERCENTAGE) value = (value * 100) / containerwidth;
   }
   else return(ObjectError(Self, ERH_GetWidth, ERR_AccessObject));

   if (Value->Type & FD_DOUBLE) Value->Double = value;
   else if (Value->Type & FD_LARGE) Value->Large = value;
   else return(ObjectError(Self, ERH_GetWidth, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

static ERROR SET_Width(struct View *Self, struct Variable *Value)
{
   if (Value->Type & FD_DOUBLE) {
      if (Value->Double <= 0) {
         DPrintF("@SetField:","[View] A width of %.2f is illegal.", Value->Double);
         return(ERR_OutOfRange);
      }
      Self->Width = Value->Double;
   }
   else if (Value->Type & FD_LARGE) {
      if (Value->Large <= 0) {
         DPrintF("@SetField:","[View] A width of %d is illegal.", Value->Large);
         return(ERR_OutOfRange);
      }
      Self->Width = Value->Large;
   }
   else return(ObjectError(Self, ERH_SetWidth, ERR_FieldTypeMismatch));

   if (Value->Type & FD_PERCENTAGE) Self->Dimensions = (Self->Dimensions & ~DMF_FIXED_WIDTH) | DMF_RELATIVE_WIDTH;
   else Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_WIDTH) | DMF_FIXED_WIDTH;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   XCoord
Short:  The horizontal position of a view.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>The horizontal position of a view can be set to an absolute or relative
coordinate by writing a value to the XCoord field.  To set a
relative/percentage based value, you must use the FD_PERCENT flag or the
value will be interpreted as fixed.  Negative values are permitted.</p>

###SEE ALSO###
Field: XOffset, YCoord

###END###

*****************************************************************************/

static ERROR GET_X(struct View *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   LONG containerwidth;
   DOUBLE width, value;

   if (AccessObject(Self->DrawableID, 4000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_Width, FT_LONG, &containerwidth);
      ReleaseObject(drawable);

      if (containerwidth <= 0) return(ERR_GetField);

      if (Self->Dimensions & DMF_FIXED_X) value = Self->XCoord;
      else if (Self->Dimensions & DMF_RELATIVE_X) {
         value = Self->XCoord * containerwidth / 100;
      }
      else if ((Self->Dimensions & (DMF_FIXED_WIDTH|DMF_RELATIVE_WIDTH)) AND
               (Self->Dimensions & (DMF_FIXED_XOFFSET|DMF_RELATIVE_XOFFSET))) {
         if (Self->Dimensions & DMF_FIXED_WIDTH) width = Self->Width;
         else width = containerwidth * Self->Width / 100;
         if (Self->Dimensions & DMF_FIXED_XOFFSET) value = containerwidth - width - Self->XOffset;
         else value = containerwidth - width - (containerwidth * Self->XOffset / 100);
      }
      else return(ObjectError(Self, ERH_GetX, ERR_InvalidDimension));

      if (Value->Type & FD_PERCENTAGE) value = (value * 100) / containerwidth;
   }
   else return(ObjectError(Self, ERH_GetX, ERR_ExclusiveDenied));

   if (Value->Type & FD_DOUBLE) Value->Double = value;
   else if (Value->Type & FD_LARGE) Value->Large = value;
   else return(ObjectError(Self, ERH_GetX, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

static ERROR SET_X(struct View *Self, struct Variable *Value)
{
   if (Value->Type & FD_DOUBLE) Self->XCoord = Value->Double;
   else if (Value->Type & FD_LARGE) Self->XCoord = Value->Large;
   else return(ObjectError(Self, ERH_SetX, ERR_FieldTypeMismatch));

   if (Value->Type & FD_PERCENTAGE) Self->Dimensions = (Self->Dimensions & ~DMF_FIXED_X) | DMF_RELATIVE_X;
   else Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_X) | DMF_FIXED_X;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   XOffset
Short:  The horizontal offset of a view.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>The XOffset has a dual purpose depending on whether or not it is set in
conjunction with an X coordinate or a Width based field.</p>

<p>If set in conjunction with an X coordinate then the view will be
drawn from that X coordinate up to the width of the container, minus the
value given in the XOffset.  This means that the width of the View is
dynamically calculated in relation to the width of the container.</p>

<p>If the XOffset field is set in conjunction with a fixed or relative width
then the view will be drawn at an X coordinate calculated from the formula
&quot;XCoord = ContainerWidth - ViewWidth - XOffset&quot;.</p>

###SEE ALSO###
Field: XCoord, Width, YOffset

###END###

*****************************************************************************/

static ERROR GET_XOffset(struct View *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   LONG containerwidth;
   DOUBLE width, value;

   if (AccessObject(Self->DrawableID, 4000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_Width, FT_LONG, &containerwidth);
      ReleaseObject(drawable);

      if (containerwidth <= 0) return(ERR_GetField);

      if (Self->Dimensions & DMF_FIXED_XOFFSET) value = Self->XOffset;
      else if (Self->Dimensions & DMF_RELATIVE_XOFFSET) {
         value = Self->XOffset * containerwidth / 100;
      }
      else if ((Self->Dimensions & (DMF_FIXED_X|DMF_RELATIVE_X)) AND
               (Self->Dimensions & (DMF_FIXED_WIDTH|DMF_RELATIVE_WIDTH))) {
         if (Self->Dimensions & DMF_FIXED_WIDTH) width = Self->Width;
         else width = containerwidth * Self->Width / 100;
         if (Self->Dimensions & DMF_FIXED_X) value = containerwidth - (Self->XCoord + width);
         else value = containerwidth - ((Self->XCoord * containerwidth / 100) + width);
      }
      else value = 0;

      if (Value->Type & FD_PERCENTAGE) value = (value * 100) / containerwidth;
   }
   else return(ObjectError(Self, ERH_GetXOffset, ERR_ExclusiveDenied));

   if (Value->Type & FD_DOUBLE) Value->Double = value;
   else if (Value->Type & FD_LARGE) Value->Large = value;
   else return(ObjectError(Self, ERH_GetXOffset, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

static ERROR SET_XOffset(struct View *Self, struct Variable *Value)
{
   if (Value->Type & FD_DOUBLE) Self->XOffset= Value->Double;
   else if (Value->Type & FD_LARGE) Self->XOffset= Value->Large;
   else return(ObjectError(Self, ERH_SetXOffset, ERR_FieldTypeMismatch));

   if (Value->Type & FD_PERCENTAGE) Self->Dimensions = (Self->Dimensions & ~DMF_FIXED_XOFFSET) | DMF_RELATIVE_XOFFSET;
   else Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_XOFFSET) | DMF_FIXED_XOFFSET;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   YCoord
Short:  The vertical position of a view.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>The vertical position of a View can be set to an absolute or relative
coordinate by writing a value to the YCoord field.  To set a
relative/percentage based value, you must use the FD_PERCENT flag or the
value will be interpreted as fixed.  Negative values are permitted.</p>

###SEE ALSO###
Field: YOffset, XCoord

###END###

*****************************************************************************/

static ERROR GET_Y(struct View *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   LONG containerheight, height;
   DOUBLE value;

   if (AccessObject(Self->DrawableID, 4000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_Height, FT_LONG, &containerheight);
      ReleaseObject(drawable);

      if (containerheight <= 0) return(ERR_GetField);

      if (Self->Dimensions & DMF_FIXED_Y) value = Self->YCoord;
      else if (Self->Dimensions & DMF_RELATIVE_Y) {
         value = Self->YCoord * containerheight / 100;
      }
      else if ((Self->Dimensions & (DMF_FIXED_HEIGHT|DMF_RELATIVE_HEIGHT)) AND
               (Self->Dimensions & (DMF_FIXED_YOFFSET|DMF_RELATIVE_YOFFSET))) {
         if (Self->Dimensions & DMF_FIXED_HEIGHT) height = Self->Height;
         else height = (containerheight * Self->Height) / 100;
         if (Self->Dimensions & DMF_FIXED_YOFFSET) value = containerheight - height - Self->YOffset;
         else value = containerheight - height - ((containerheight * Self->YOffset) / 100);
      }
      else value = 0;

      if (Value->Type & FD_PERCENTAGE) value = (value * 100) / containerheight;
   }
   else return(ObjectError(Self, ERH_GetY, ERR_AccessObject));

   if (Value->Type & FD_DOUBLE) Value->Double = value;
   else if (Value->Type & FD_LARGE) Value->Large = value;
   else return(ObjectError(Self, ERH_GetY, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

static ERROR SET_Y(struct View *Self, struct Variable *Value)
{
   if (Value->Type & FD_DOUBLE) Self->YCoord = Value->Double;
   else if (Value->Type & FD_LARGE) Self->YCoord = Value->Large;
   else return(ObjectError(Self, ERH_SetY, ERR_FieldTypeMismatch));

   if (Value->Type & FD_PERCENTAGE) Self->Dimensions = (Self->Dimensions & ~DMF_FIXED_Y) | DMF_RELATIVE_Y;
   else Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_Y) | DMF_FIXED_Y;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   YOffset
Short:  The vertical offset of a view.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>The YOffset has a dual purpose depending on whether or not it is set in
conjunction with a Y coordinate or a Height based field.</p>

<p>If set in conjunction with a Y coordinate then the view will be drawn
from that Y coordinate up to the height of the container, minus the value
given in the YOffset.  This means that the height of the view is
dynamically calculated in relation to the height of the container.</p>

<p>If the YOffset field is set in conjunction with a fixed or relative height
then the view will be drawn at a Y coordinate calculated from the formula
&quot;YCoord = ContainerHeight - ViewHeight - YOffset&quot;.</p>

###SEE ALSO###
Field: YCoord, Height, XOffset

###END###

*****************************************************************************/

static ERROR GET_YOffset(struct View *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   LONG containerheight;
   DOUBLE value, height;

   value = 0;
   if (AccessObject(Self->DrawableID, 4000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_Height, FT_LONG, &containerheight);
      ReleaseObject(drawable);

      if (containerheight <= 0) return(ERR_GetField);

      if (Self->Dimensions & DMF_FIXED_YOFFSET) value = Self->YOffset;
      else if (Self->Dimensions & DMF_RELATIVE_YOFFSET) {
         value = Self->YOffset * containerheight / 100;
      }
      else if ((Self->Dimensions & (DMF_FIXED_Y|DMF_RELATIVE_Y)) AND
               (Self->Dimensions & (DMF_FIXED_HEIGHT|DMF_RELATIVE_HEIGHT))) {
         if (Self->Dimensions & DMF_FIXED_HEIGHT) height = Self->Height;
         else height = containerheight * Self->Height / 100;
         if (Self->Dimensions & DMF_FIXED_Y) value = containerheight - (Self->YCoord + height);
         else value = containerheight - ((Self->YCoord * containerheight / 100) + height);
      }
      else value = 0;

      if (Value->Type & FD_PERCENTAGE) value = (value * 100) / containerheight;
   }
   else return(ObjectError(Self, ERH_GetYOffset, ERR_ExclusiveDenied));

   if (Value->Type & FD_DOUBLE) Value->Double = value;
   else if (Value->Type & FD_LARGE) Value->Large = value;
   else return(ObjectError(Self, ERH_GetYOffset, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

static ERROR SET_YOffset(struct View *Self, struct Variable *Value)
{
   if (Value->Type & FD_DOUBLE) Self->YOffset= Value->Double;
   else if (Value->Type & FD_LARGE) Self->YOffset= Value->Large;
   else return(ObjectError(Self, ERH_SetYOffset, ERR_FieldTypeMismatch));

   if (Value->Type & FD_PERCENTAGE) Self->Dimensions = (Self->Dimensions & ~DMF_FIXED_YOFFSET) | DMF_RELATIVE_YOFFSET;
   else Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_YOFFSET) | DMF_FIXED_YOFFSET;
   return(ERR_Okay);
}

/*****************************************************************************
** Internal: CaclulateHScroll()
*/

static ERROR CalculateHScroll(struct View *Self)
{
   struct mtUpdateScroll scroll;

   //DPrintF("CalculateHScroll()","[View:%d] Page: %d, View: %d", Self->Head.UniqueID, Self->AreaX + Self->PageWidth, Self->AreaWidth);

   if (!Self->HScroll) return(ERR_Okay);
   if (Self->PageWidth < 0) Self->PageWidth = 0;

   scroll.ViewSize = -1;
   scroll.PageSize = Self->AreaX + Self->PageWidth;
   scroll.Position = -Self->XPosition;
   scroll.Unit     = 16;
   return(ActionMsg(MT_UpdateScroll, Self->HScroll, &scroll));
}

/*****************************************************************************
** Internal: CaclulateVScroll()
*/

static ERROR CalculateVScroll(struct View *Self)
{
   struct mtUpdateScroll scroll;

   //DPrintF("CalculateVScroll()","[View:%d] Page: %d, View: %d", Self->Head.UniqueID, Self->AreaY + Self->PageHeight, Self->AreaHeight);

   if (!Self->VScroll) return(ERR_Okay);
   if (Self->PageHeight < 0) Self->PageHeight = 0;

   scroll.ViewSize = -1;
   scroll.PageSize = Self->AreaY + Self->PageHeight;
   scroll.Position = -Self->YPosition;
   scroll.Unit     = Self->Font->MaxHeight;
   return(ActionMsg(MT_UpdateScroll, Self->VScroll, &scroll));
}

/*****************************************************************************
** Internal: CheckPointerCursor()
*/

static void CheckPointerCursor(struct View *Self, WORD XCoord, WORD YCoord)
{
   struct Column *col;
   WORD cx;

   if (Self->Style IS VIEW_COLUMN) {
      if ((YCoord >= Self->AreaY) AND (YCoord < Self->AreaY + Self->ColumnHeight)) {
         cx = 0;
         for (col=Self->Columns; col; col=col->Next) {
            cx += col->Width;

            if ((XCoord - Self->XPosition >= cx-3) AND (XCoord - Self->XPosition < cx+3)) {
               if (!Self->PointerLocked) {
                  if (SetCursor(Self->DrawableID, NULL, PTR_SPLIT_HORIZONTAL, NULL, Self->Head.UniqueID, NULL) IS ERR_Okay) {
                     Self->PointerLocked = PTR_SPLIT_HORIZONTAL;
                  }
               }
               return;
            }
         }
      }
   }

   if ((Self->PointerLocked) AND (!Self->ColumnResize)) {
      /* We have the pointer locked and the cursor is out of bounds.
      ** Assuming that the user is not performing a column resize,
      ** restore the cursor back to its normal state.
      */
      RestoreCursor(PTR_DEFAULT, Self->Head.UniqueID);
      Self->PointerLocked = NULL;
   }
}

/*****************************************************************************
** Internal: GetXMLAttribute()
*/

static STRING GetXMLAttribute(struct XMLTag *Tag, STRING Attribute)
{
   WORD i;

   for (i=0; i < Tag->TotalAttributes; i++) {
      if (StrCompare(Tag->Attributes[i].Name, Attribute, 0, STR_MATCHLENGTH) IS ERR_Okay)  {
         return(Tag->Attributes[i].Value);
      }
   }
   return(NULL);
}

/*****************************************************************************
** Internal: ArrangeItems()
*/

static void ArrangeItems(struct View *Self)
{
   struct mtCalcStringWidth calc;
   struct XMLTag **xmllist, *tag, *scan;
   struct ViewNode *node, *scan_node;
   struct Column *col;
   LONG x, y;
   UBYTE buffer[400];
   WORD index, columngap, linewidth, columncount, hbar;

   #ifdef DEBUG
      DPrintF("~ArrangeItems()","[View:%d]", Self->Head.UniqueID);
   #endif

   Self->PageWidth  = 0;
   Self->PageHeight = 0;
   Self->IconWidth  = 0;
   Self->LineHeight = 0;

   if (Self->Style IS VIEW_COLUMN) Self->ColumnHeight = Self->Font->MaxHeight + 6;
   else Self->ColumnHeight = 0;

   if ((GetField(Self->XML, FID_Tags, FT_POINTER, &xmllist) != ERR_Okay) OR (!xmllist)) {
      CalculateVScroll(Self);
      CalculateHScroll(Self);
      #ifdef DEBUG
         StepBack();
      #endif
      return;
   }

   if (Self->Style IS VIEW_TREE) {

   }
   else if ((Self->Style IS VIEW_LIST) OR (Self->Style IS VIEW_LONGLIST)) {

      columncount = 0;
      columngap = 20;

      if (!(Self->Flags & VWF_NOICONS)) {
         Self->IconWidth  = 16 + 4; /* Add extra pixels to create a gap between the icon and the text */
         Self->LineHeight = 16;     /* Line height must meet the icon size, at a minimum */
      }

      /*** Determine the number of pixels assigned to each line ***/

      if (Self->LineHeight < Self->Font->MaxHeight) {
         Self->LineHeight = Self->Font->MaxHeight;
         if (Self->VSpacing) Self->LineHeight += Self->VSpacing;
         else Self->LineHeight += 4;
      }
      else {
         if (Self->VSpacing) Self->LineHeight += Self->VSpacing;
         else Self->LineHeight += 1;
      }

      if ((Self->Style IS VIEW_LIST) AND (Self->HScroll)) {
         /* Horizontal scrollbar height compensation, helps avoid the
         ** vertical scrollbar from having to be used in list mode.
         */
         hbar = 20;
      }
      else hbar = 0;

      x = 0;
      y = Self->TopMargin;
      linewidth = 0;
      tag = xmllist[0];
      for (index=0; tag; index++) {
         node = tag->Private;

         ActionTags(MT_GetXMLContent, Self->XML, tag->Index, buffer, sizeof(buffer));

         calc.Width = 0;
         if (buffer[0]) {
            calc.String = buffer;
            calc.Length = -1;
            Action(MT_CalcStringWidth, Self->Font, &calc);
         }

         if (Self->IconWidth + calc.Width + columngap > linewidth) {
            linewidth = Self->IconWidth + calc.Width + columngap;
         }

         if (linewidth > Self->MaxItemWidth) linewidth = Self->MaxItemWidth;

         node->XCoord = x;
         node->YCoord = y;
         node->Width  = 0; /* We'll set this later */
         node->Height = Self->LineHeight;

         if (y IS Self->AreaY) node->Flags |= NODE_NEWCOLUMN;
         else node->Flags &= ~NODE_NEWCOLUMN;

         y += Self->LineHeight;

         if ((!tag->Next) OR ((Self->Style IS VIEW_LIST) AND (y > Self->AreaY) AND (y + Self->LineHeight > Self->AreaHeight - hbar))) {
            /*** We are about to go to a new column, or this is the end of all columns ***/

            if ((!tag->Next) AND (columncount < 1)) {
               /* If there are no more tags and we know there is only one column,
               ** all nodes will use the entire width of view.
               */
               for (scan=tag; scan; scan=scan->Prev) {
                  scan_node = (struct ViewNode *)scan->Private;
                  if (scan_node->Width) break;
                  scan_node->Width = Self->AreaWidth;
               }
            }
            else {
               for (scan=tag; scan; scan=scan->Prev) {
                  scan_node = (struct ViewNode *)scan->Private;
                  if (scan_node->Width) break;
                  scan_node->Width = linewidth;
               }
            }

            y = Self->AreaY;
            x += linewidth;
            linewidth = 0;
            columncount++;
         }

         if (Self->Style IS VIEW_LONGLIST) {
            if (linewidth > Self->PageWidth) Self->PageWidth = linewidth;
         }
         else if (node->XCoord + linewidth > Self->PageWidth) Self->PageWidth = node->XCoord + linewidth;

         if (node->YCoord + node->Height > Self->PageHeight) Self->PageHeight = node->YCoord + node->Height;

         tag = tag->Next;
      }
   }
   else if (Self->Style IS VIEW_ICON) {

   }
   else if (Self->Style IS VIEW_COLUMN) {


      if (!(Self->Flags & VWF_NOICONS)) {
         Self->IconWidth  = 16 + 4; /* Add extra pixels to create a gap between the icon and the text */
         Self->LineHeight = 16;
      }

      x = 0; //Self->AreaX;
      y = Self->ColumnHeight; //Self->AreaY + Self->ColumnHeight;
      columncount = 0;
      columngap = 20;

      /*** Determine the number of pixels assigned to each line ***/

      if (Self->LineHeight < Self->Font->MaxHeight) {
         Self->LineHeight = Self->Font->MaxHeight;
         if (Self->VSpacing) Self->LineHeight += Self->VSpacing;
         else Self->LineHeight += 4;
      }
      else {
         if (Self->VSpacing) Self->LineHeight += Self->VSpacing;
         else Self->LineHeight += 1;
      }

      /*** Calculate the width of the scrollable page ***/

      for (Self->PageWidth=0, col=Self->Columns; col; col=col->Next) Self->PageWidth += col->Width;

      node = 0;
      linewidth = 0;
      tag = xmllist[0];
      for (index=0; tag; index++) {
         node = tag->Private;

         ActionTags(MT_GetXMLContent, Self->XML, tag->Index, buffer, sizeof(buffer));

         calc.Width = 0;
         if (buffer[0]) {
            calc.String = buffer;
            calc.Length = -1;
            Action(MT_CalcStringWidth, Self->Font, &calc);
         }

         if (Self->IconWidth + calc.Width + columngap > linewidth) {
            linewidth = Self->IconWidth + calc.Width + columngap;
         }

         node->XCoord = x;
         node->YCoord = y;
         if (Self->AreaWidth > Self->PageWidth) node->Width = Self->AreaWidth;
         else node->Width = Self->PageWidth; /* In column mode the entire breadth of the line is used for each item */
         node->Height = Self->LineHeight;

         y += Self->LineHeight;

         tag = tag->Next;
      }

      /* Calculate the height of the scrollable page (does not
      ** include the column buttons at the top, as they are not
      ** included in the scroll proces).
      */

      if (node) Self->PageHeight = node->YCoord + node->Height;
      else Self->PageHeight = 0;
   }

   /*** Recalculate scroll bars, based on the new PageWidth and PageHeight values ***/

   CalculateVScroll(Self);
   CalculateHScroll(Self);

   #ifdef DEBUG
      StepBack();
   #endif
}

/*****************************************************************************
** Internal: SortItems()
*/

static ERROR SortItems(struct View *Self)
{
   struct mtSortXML sort;
   struct Column *col;
   struct XMLSort sortlist[2] = { { "Default", NULL, XSF_CHECKSORT|XSF_REPORTSORTED }, { NULL, NULL, NULL } };
   WORD colindex, i;
   ERROR error;

   DPrintF("~6SortItems()","Column: %d", Self->Sort[0]);

   colindex = Self->Sort[0];
   if (colindex < 0) { colindex = -colindex; sortlist[0].Flags |= XSF_DESCENDING; }
   colindex--;

   for (i=0, col=Self->Columns; (col) AND (i < colindex); col=col->Next, i++);

   if (!col) {
      col = Self->Columns;
      Self->Sort[0] = 1;
   }

   /* Ask the XML object to resort the XML.  This will sort on tag content
   ** by default, although the developer can override this by passing the
   ** 'sort' attribute amongst the tags to affect/improve sorting behaviour.
   */

   if (col != Self->Columns) sortlist[0].TagName = col->Name;
   sortlist[0].Attribute = NULL; /* Will sort on content, not an attribute */
   sort.Index = -1;
   sort.List  = sortlist;
   error = Action(MT_SortXML, Self->XML, &sort);

   StepBack();
   return(error);
}

/*****************************************************************************
** Internal: DrawTree()
*/

static void DrawTree(struct View *Self, struct Bitmap *Bitmap, struct XMLTag *Tag)
{
   struct ViewNode *node;

   while (Tag) {
      node = Tag->Private;
      /*** Draw the opening box ***/

      ActionTags(MT_DrawRGBRectangle, Bitmap, node->XCoord, node->YCoord, Self->IconWidth, Self->LineHeight, 0, 0, 0, FALSE);

      /*** Draw connection lines ***/

      /*** Draw the icon ***/

      /*** Draw the text ***/



      if ((Tag->Child) AND (Tag->Child->Attributes) AND
          (StrCompare(Tag->Attributes[0].Name, Tag->Child->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay)){
         DrawTree(Self, Bitmap, Tag->Child);
      }
      Tag = Tag->Next;
   }
}

/*****************************************************************************
** Internal: DrawButton()
*/

static void DrawButton(struct View *Self, struct Bitmap *Bitmap, WORD X, WORD Y, WORD Width, WORD Height)
{
   Height--;

   ActionTags(MT_DrawRGBRectangle, Bitmap, X, Y,        Width, 1,   Self->ColHighlight.Red, Self->ColHighlight.Green, Self->ColHighlight.Blue, TRUE); /* Highlight (Top) */
   ActionTags(MT_DrawRGBRectangle, Bitmap, X, Y,        1, Height,    Self->ColHighlight.Red, Self->ColHighlight.Green, Self->ColHighlight.Blue, TRUE);   /* Highlight (Left) */
   ActionTags(MT_DrawRGBRectangle, Bitmap, X+Width-1,   Y, 1, Height, Self->ColShadow.Red, Self->ColShadow.Green, Self->ColShadow.Blue, TRUE); /* Shadow (Right) */
   ActionTags(MT_DrawRGBRectangle, Bitmap, X, Y+Height, Width, 1,     Self->ColShadow.Red, Self->ColShadow.Green, Self->ColShadow.Blue, TRUE); /* Shadow (Bottom) */

   if (Self->ColThickness > 1) {
      LONG r, g, b;
      X++;
      Y++;
      Width  -= 2;
      Height -= 2;
      r = Self->ColBackground.Red   + ((Self->ColHighlight.Red - Self->ColBackground.Red)>>1);
      g = Self->ColBackground.Green + ((Self->ColHighlight.Green - Self->ColBackground.Green)>>1);
      b = Self->ColBackground.Blue  + ((Self->ColHighlight.Blue - Self->ColBackground.Blue)>>1);
      ActionTags(MT_DrawRGBRectangle, Bitmap, X, Y, Width-2, 1, r, g, b, TRUE); /* Highlight (Inner-Top) */
      ActionTags(MT_DrawRGBRectangle, Bitmap, X, Y, 1, Height, r, g, b, TRUE); /* Highlight (Inner-Left) */

      r = Self->ColShadow.Red   + ((Self->ColBackground.Red - Self->ColShadow.Red)>>1);
      g = Self->ColShadow.Green + ((Self->ColBackground.Green - Self->ColShadow.Green)>>1);
      b = Self->ColShadow.Blue  + ((Self->ColBackground.Blue - Self->ColShadow.Blue)>>1);
      //ActionTags(MT_DrawRGBRectangle, Bitmap, Width-1, Y, 1, Height, r, g, b, TRUE); /* Shadow (Inner-Right) */
      ActionTags(MT_DrawRGBRectangle, Bitmap, X, Y+Height, Width, 1, r, g, b, TRUE);  /* Shadow (Inner-Bottom) */
   }
}

/*****************************************************************************
** Internal: DrawView()
*/

static void DrawView(struct View *Self, struct Render *Drawable, struct Bitmap *Bitmap)
{
   struct XMLTag **xmllist, *tag;
   struct ViewNode *node;
   struct Font *font;
   LONG clipleft, cliptop, clipbottom, clipright;
   LONG saveleft, savetop, savebottom, saveright;
   UBYTE buffer[400];
   WORD i;

   if (Self->Flags & VWF_INVISIBLE) return;

   font = Self->Font;

   saveleft   = Bitmap->ClipLeft;
   savetop    = Bitmap->ClipTop;
   saveright  = Bitmap->ClipRight;
   savebottom = Bitmap->ClipBottom;

   if (Self->AreaX > Bitmap->ClipLeft) Bitmap->ClipLeft = Self->AreaX;
   if (Self->AreaY > Bitmap->ClipTop)  Bitmap->ClipTop = Self->AreaY;
   if (Self->AreaX + Self->AreaWidth < Bitmap->ClipRight)   Bitmap->ClipRight = Self->AreaX + Self->AreaWidth;
   if (Self->AreaY + Self->AreaHeight < Bitmap->ClipBottom) Bitmap->ClipBottom = Self->AreaY + Self->AreaHeight;

   clipleft   = Bitmap->ClipLeft;
   cliptop    = Bitmap->ClipTop;
   clipright  = Bitmap->ClipRight;
   clipbottom = Bitmap->ClipBottom;

   if (Self->Style IS VIEW_TREE) {
      if ((GetField(Self->XML, FID_Tags, FT_POINTER, &xmllist) != ERR_Okay) OR (!xmllist)) goto exit;

      DrawTree(Self, Bitmap, xmllist[0]);
   }
   else if ((Self->Style IS VIEW_LIST) OR (Self->Style IS VIEW_LONGLIST)) {
      LONG x, y;

      if ((GetField(Self->XML, FID_Tags, FT_POINTER, &xmllist) != ERR_Okay) OR (!xmllist)) goto exit;

      font->Bitmap = Bitmap;

      for (tag=xmllist[0]; tag; tag = tag->Next) {
         if (!(node = tag->Private)) continue;

         x = Self->AreaX + node->XCoord + Self->XPosition;
         y = Self->AreaY + node->YCoord + Self->YPosition;

         if (x + node->Width <= Bitmap->ClipLeft) continue;
         if (y + node->Height <= Bitmap->ClipTop) continue;
         if (x >= Bitmap->ClipRight) break;
         if (y >= Bitmap->ClipBottom) break;

         /*** If the item is selected, draw a highlight for it ***/

         if ((node->Flags & NODE_SELECTED) AND (!(Drawable->Flags & RNF_DISABLED))) {
            ActionTags(MT_DrawRGBRectangle, Bitmap, x, y, node->Width, Self->LineHeight,
               Self->SelectRGB.Red, Self->SelectRGB.Green, Self->SelectRGB.Blue, TRUE);
         }

         /*** Draw the icon on the left ***/

         if (node->Icon) {
            node->Icon->Bitmap->Flags |= BMF_ALPHABLEND;
            if (Drawable->Flags & RNF_DISABLED) node->Icon->Bitmap->Opacity = 128;
            ActionTags(MT_BlitArea, node->Icon->Bitmap, Bitmap, NULL, 0, 0,
               node->Icon->Bitmap->Width, node->Icon->Bitmap->Height,
               x + Self->LeftMargin + ((Self->IconWidth - node->Icon->Bitmap->Width)/2),
               y + ((Self->LineHeight - node->Icon->Bitmap->Height)/2));
            if (Drawable->Flags & RNF_DISABLED) node->Icon->Bitmap->Opacity = 255;
         }

         /*** Draw the text alongside the icon ***/

         ActionTags(MT_GetXMLContent, Self->XML, tag->Index, buffer, sizeof(buffer));

         if (buffer[0]) {
            font->XCoord = x + Self->LeftMargin + Self->IconWidth + 2;
            font->YCoord = y + ((Self->LineHeight - font->Height)/2);
            font->WrapWidth = x + node->Width;

            if ((tag->Index IS Self->HighlightTag) AND (!(Self->Buttons)) AND (!(Drawable->Flags & RNF_DISABLED))) {
               font->ColourRGB->Red   = Self->HighlightRGB.Red;
               font->ColourRGB->Green = Self->HighlightRGB.Green;
               font->ColourRGB->Blue  = Self->HighlightRGB.Blue;
            }
            else {
               font->ColourRGB->Red   = node->FontRGB.Red;
               font->ColourRGB->Green = node->FontRGB.Green;
               font->ColourRGB->Blue  = node->FontRGB.Blue;
            }
            SetFields(font, FID_String|TSTR, buffer, TAGEND);

            if (Drawable->Flags & RNF_DISABLED) SetFields(font, FID_Opacity|TLONG, 50, TAGEND);

            Action(AC_Draw, font, NULL);

            if (Drawable->Flags & RNF_DISABLED) SetFields(font, FID_Opacity|TLONG, 100, TAGEND);
         }
      }
   }
   else if (Self->Style IS VIEW_ICON) {

   }
   else if (Self->Style IS VIEW_COLUMN) {
      struct Column *col;
      struct XMLTag *child;
      WORD colindex, ax, ay, ah;
      LONG x, y;

      font->Bitmap = Bitmap;

      /*** Draw column buttons at the top ***/

      x = Self->AreaX + Self->XPosition;
      y = Self->AreaY + Self->YPosition;

      /*** Draw complete button background in one shot ***/

      ActionTags(MT_DrawRGBRectangle, Bitmap, Self->AreaX, Self->AreaY, Self->AreaWidth, Self->ColumnHeight,
         Self->ColBackground.Red, Self->ColBackground.Green, Self->ColBackground.Blue, TRUE);

      colindex = 0;
      for (col=Self->Columns; col; col=col->Next, colindex++) {
         /*** Adjust clipping to match that of the current column ***/

         if (x > clipleft) Bitmap->ClipLeft = x;
         if ((x + col->Width) < clipright) Bitmap->ClipRight = x + col->Width;

         /*** Draw button ***/

         if (Self->Flags & VWF_DRAWTABLE) {
            ActionTags(MT_DrawRGBRectangle, Bitmap, x+col->Width-1, Self->AreaY, 1, Self->ColumnHeight-1, Self->ColShadow.Red, Self->ColShadow.Green, Self->ColShadow.Blue, TRUE);
         }
         else DrawButton(Self, Bitmap, x, Self->AreaY, col->Width, Self->ColumnHeight);

         /*** Draw sort arrow ***/

         font->XCoord = x + 4;
         font->YCoord = Self->AreaY + ((Self->ColumnHeight - font->Height)/2);
         font->AlignWidth = col->Width - 10;
         font->WrapWidth = x + col->Width;
         font->ColourRGB->Red   = 0;
         font->ColourRGB->Green = 0;
         font->ColourRGB->Blue  = 0;

         if (col->Width >= 20) {
            if (Self->Sort[0] IS -(colindex+1)) {
               /*** Up arrow (descending sort) ***/
               ah = 5;
               ax = x + col->Width - (ah*2) - 2;
               ay = font->YCoord + ((font->Height - ah)/2);
               for (i=0; i < ah; i++) {
                  ActionTags(MT_DrawRGBLine, Bitmap, ax, ay, ax + (i * 2), ay, 80, 80, 160);
                  ax--;
                  ay++;
               }
               font->AlignWidth -= (ah * 2) + 6;
            }
            else if (Self->Sort[0] IS (colindex+1)) {
               /*** Down arrow (ascending sort) ***/
               ah = 5;
               ax = x + col->Width - (ah*2) - 2;
               ay = font->YCoord + ((font->Height - ah)/2) + ah;
               for (i=0; i < ah; i++) {
                  ActionTags(MT_DrawRGBLine, Bitmap, ax, ay, ax + (i * 2), ay, 80, 80, 160);
                  ax--;
                  ay--;
               }

               font->AlignWidth -= (ah * 2) + 6;
            }
         }

         /*** Draw text inside the button ***/

         if (col->Flags & CF_RIGHTALIGN) {
            font->AlignFlags |= ALIGN_RIGHT;
         }

         SetFields(font, FID_String|TSTR, col->Text, TAGEND);
         //if (Drawable->Flags & RNF_DISABLED) SetFields(font, FID_Opacity|TLONG, 50, TAGEND);
         Action(AC_Draw, font, NULL);
         //if (Drawable->Flags & RNF_DISABLED) SetFields(font, FID_Opacity|TLONG, 10, TAGEND);

         font->AlignFlags = NULL;

         /*** Draw background for entire vertical column, if required ***/

         if (col->Flags & CF_COLOUR) {
            ActionTags(MT_DrawRGBRectangle, Bitmap, x, Self->AreaY + Self->ColumnHeight,
               col->Width, Self->AreaHeight - Self->AreaY - Self->ColumnHeight,
               col->Colour.Red, col->Colour.Green, col->Colour.Blue, TRUE);
         }

         /*** Draw hairlines between each column if this option is on ***/

         if (Self->Flags & (VWF_HAIRLINES|VWF_DRAWTABLE)) {
            if ((col->Next) OR (x+col->Width < Self->AreaWidth)) {
               ActionTags(MT_DrawRGBRectangle, Bitmap, x+col->Width-1, Self->AreaY + Self->ColumnHeight,
                  1, Self->AreaHeight - (Self->AreaY + Self->ColumnHeight),
                  Self->HairlineRGB.Red, Self->HairlineRGB.Green, Self->HairlineRGB.Blue, TRUE);
            }
         }

         Bitmap->ClipLeft  = clipleft;
         Bitmap->ClipRight = clipright;

         x += col->Width;
      }

      /*** Draw an empty button to fill any left-over space to the right of the columns ***/

      if ((x < Self->AreaWidth) AND (!(Self->Flags & VWF_DRAWTABLE))) {
         DrawButton(Self, Bitmap, x, Self->AreaY, 16000, Self->ColumnHeight);
      }

      if ((GetField(Self->XML, FID_Tags, FT_POINTER, &xmllist) != ERR_Okay) OR (!xmllist)) goto exit;

      if (Self->AreaY + Self->ColumnHeight > cliptop) Bitmap->ClipTop = Self->AreaY + Self->ColumnHeight;

      /*** Draw the items ***/

      for (tag=xmllist[0]; tag; tag = tag->Next) {
         node = tag->Private;

         x = Self->AreaX + node->XCoord + Self->XPosition;
         y = Self->AreaY + node->YCoord + Self->YPosition;

         if (x + node->Width <= clipleft) break;
         if (y + node->Height <= cliptop) continue;
         if (x >= clipright) break;
         if (y >= clipbottom) break;

         Bitmap->ClipLeft = clipleft;
         Bitmap->ClipRight = clipright;

         /*** If the item is selected, draw a highlight for it ***/

         if ((node->Flags & NODE_SELECTED) AND (!(Drawable->Flags & RNF_DISABLED))) {
            ActionTags(MT_DrawRGBRectangle, Bitmap, x, y,
               8192, Self->LineHeight, Self->SelectRGB.Red, Self->SelectRGB.Green, Self->SelectRGB.Blue, TRUE);
         }

         if ((tag->Index IS Self->HighlightTag) AND (!(Self->Buttons)) AND (!(Drawable->Flags & RNF_DISABLED))) {
            font->ColourRGB->Red   = Self->HighlightRGB.Red;
            font->ColourRGB->Green = Self->HighlightRGB.Green;
            font->ColourRGB->Blue  = Self->HighlightRGB.Blue;
         }
         else {
            font->ColourRGB->Red   = node->FontRGB.Red;
            font->ColourRGB->Green = node->FontRGB.Green;
            font->ColourRGB->Blue  = node->FontRGB.Blue;
         }

         for (col=Self->Columns; col; x += col->Width, col=col->Next) {
            /*** Adjust clipping to match that of the current column ***/

            if (x > clipleft) Bitmap->ClipLeft = x;
            else Bitmap->ClipLeft = clipleft;

            if ((x + col->Width) < clipright) Bitmap->ClipRight = x + col->Width;
            else Bitmap->ClipRight = clipright;

            node = NULL;
            buffer[0] = 0;
            if (StrCompare("Default", col->Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
               ActionTags(MT_GetXMLContent, Self->XML, tag->Index, buffer, sizeof(buffer));
               node = tag->Private;
            }
            else {
               /* Scan for the tag that matches that set against the column.  If it
               ** doesn't exist then we'll print nothing in this column.
               **
               ** Column data can either exist in a child tag first, or a tag attribute
               ** if not matching child tags are available.
               */

               for (child=tag->Child; child; child=child->Next) {
                  if (StrCompare(child->Attributes[0].Name, col->Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
                     ActionTags(MT_GetXMLContent, Self->XML, child->Index, buffer, sizeof(buffer));
                     node = child->Private;
                     break;
                  }
               }

               if (!node) {
                  for (i=0; i < tag->TotalAttributes; i++) {
                     if (StrCompare(tag->Attributes[i].Name, col->Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
                        StrCopy(tag->Attributes[i].Value, buffer, sizeof(buffer));
                        node = tag->Private;
                        break;
                     }
                  }
               }
            }

            if (!node) continue;

            /*** Draw the icon for this column ***/

            if ((col->Flags & CF_SHOWICONS) AND (node->Icon)) {
               node->Icon->Bitmap->Flags |= BMF_ALPHABLEND;
               if (Drawable->Flags & RNF_DISABLED) node->Icon->Bitmap->Opacity = 128;

               ActionTags(MT_BlitArea, node->Icon->Bitmap, Bitmap, NULL, 0, 0,
                  node->Icon->Bitmap->Width, node->Icon->Bitmap->Height,
                  x + 2 + ((Self->IconWidth - node->Icon->Bitmap->Width)/2),
                  y + ((Self->LineHeight - node->Icon->Bitmap->Height)/2));

               if (Drawable->Flags & RNF_DISABLED) node->Icon->Bitmap->Opacity = 255;
            }

            if (buffer[0]) {
               if (col->Type IS CT_DATE) {
                  struct Time time;
                  UBYTE str[6];
                  WORD j;

                  j = 0;
                  i = 0;
                  for (i=0; (buffer[j]) AND (i < 4); i++) str[i] = buffer[j++]; str[i] = 0;
                  if ((time.Year = StrToInt(str))) {
                     for (i=0; (buffer[j]) AND (i < 2); i++) str[i] = buffer[j++]; str[i] = 0;
                     time.Month = StrToInt(str);
                     for (i=0; (buffer[j]) AND (i < 2); i++) str[i] = buffer[j++]; str[i] = 0;
                     time.Day = StrToInt(str);
                     while ((buffer[j]) AND (buffer[j] <= 0x20)) j++;
                     for (i=0; (buffer[j]) AND (i < 2); i++) str[i] = buffer[j++]; str[i] = 0;
                     time.Hour = StrToInt(str);
                     if (buffer[j] IS ':') j++;
                     for (i=0; (buffer[j]) AND (i < 2); i++) str[i] = buffer[j++]; str[i] = 0;
                     time.Minute = StrToInt(str);
                     if (buffer[j] IS ':') j++;
                     for (i=0; (buffer[j]) AND (i < 2); i++) str[i] = buffer[j++]; str[i] = 0;
                     time.Second = StrToInt(str);

                     StrFormatDate(buffer, sizeof(buffer), Self->DateFormat, &time);
                  }
                  else buffer[0] = 0;
               }
               else if (col->Type IS CT_BYTESIZE) {
                  DOUBLE number;
                  number = StrToFloat(buffer);
                  if (number < 1024.0) StrFormat(buffer, sizeof(buffer), "%d", (LONG)number);
                  else if (number < 1048576.0) StrFormat(buffer, sizeof(buffer), "%.0f KB", number / 1024.0);
                  else if (number < 1073741824.0) {
                     number /= 1048576.0;
                     if (number >= 10.0) StrFormat(buffer, sizeof(buffer), "%.0f MB", number);
                     else StrFormat(buffer, sizeof(buffer), "%.1f MB", number);
                  }
                  else StrFormat(buffer, sizeof(buffer), "%.1f GB", number / 1073741824.0);
               }

               if (col->Flags & CF_RIGHTALIGN) {
                  font->AlignFlags |= ALIGN_RIGHT;
                  font->AlignWidth = col->Width - 8;
               }

               if ((node->Icon) AND (col->Flags & CF_SHOWICONS)) font->XCoord = x + Self->IconWidth + 4;
               else font->XCoord = x + 4;

               font->YCoord = Self->AreaY + Self->YPosition + ((struct ViewNode *)(tag->Private))->YCoord + ((Self->LineHeight - font->Height)/2);
               font->WrapWidth = x + col->Width;

               SetFields(font, FID_String|TSTR, buffer, TAGEND);
               if (Drawable->Flags & RNF_DISABLED) SetFields(font, FID_Opacity|TLONG, 50, TAGEND);
               Action(AC_Draw, font, NULL);
               if (Drawable->Flags & RNF_DISABLED) SetFields(font, FID_Opacity|TLONG, 100, TAGEND);

               font->AlignFlags = NULL;
            } /* if (buffer[0]) */
         } /* for (col=Self->Columns; col; x += col->Width, col=col->Next) */
      } /* for (tag=xmllist[0]; tag; tag = tag->Next) */

      Bitmap->ClipTop    = cliptop;
      Bitmap->ClipBottom = clipbottom;
      Bitmap->ClipLeft   = clipleft;
      Bitmap->ClipRight  = clipright;

      if (Self->Flags & VWF_DRAWTABLE) {
         /*** Draw a border around the button background if in table mode ***/
         ActionTags(MT_DrawRGBRectangle, Bitmap, Self->AreaX, Self->AreaY, Self->AreaWidth, Self->ColumnHeight,
            Self->ColShadow.Red, Self->ColShadow.Green, Self->ColShadow.Blue, FALSE);

         ActionTags(MT_DrawRGBRectangle, Bitmap, Self->AreaX, Self->AreaY, Self->AreaWidth, Self->AreaHeight,
            Self->ColShadow.Red, Self->ColShadow.Green, Self->ColShadow.Blue, FALSE);
      }

#if 0
      /*** Don't like this, looks better when the hairlines are behind selected items ***/
      /*** Draw hairlines between each column if this option is on ***/

      if (Self->Flags & (VWF_HAIRLINES|VWF_DRAWTABLE)) {
         Bitmap->ClipLeft = clipleft;
         Bitmap->ClipRight = clipright;

         x = Self->AreaX + Self->XPosition;
         for (col=Self->Columns; col; col=col->Next) {
            if ((col->Next) OR (x+col->Width < Self->AreaWidth)) {
               ActionTags(MT_DrawRGBRectangle, Bitmap, x+col->Width-1, Self->AreaY + Self->ColumnHeight,
                  1, Self->AreaHeight - Self->AreaY - Self->ColumnHeight,
                  Self->HairlineRGB.Red, Self->HairlineRGB.Green, Self->HairlineRGB.Blue, TRUE);
            }
            x += col->Width;
         }
      }
#endif
   }

//   ActionTags(MT_DrawRectangle, Bitmap, Bitmap->ClipLeft, Bitmap->ClipTop, Bitmap->ClipRight - Bitmap->ClipLeft, Bitmap->ClipBottom - Bitmap->ClipTop, PackPixel(Bitmap, RandomNumber(255), 0x80 + RandomNumber(125), RandomNumber(255)), TRUE);

exit:
   Bitmap->ClipLeft   = saveleft;
   Bitmap->ClipTop    = savetop;
   Bitmap->ClipRight  = saveright;
   Bitmap->ClipBottom = savebottom;
}

/******************************************************************************
** Function: CheckItemVisible()
*/

static void CheckItemVisible(struct View *Self, struct XMLTag *Tag)
{
   struct ViewNode *node;
   DOUBLE scrollx, scrolly;
   LONG itemx, itemy;

   if (!Tag) return;

   node = Tag->Private;

   scrollx = -Self->XPosition;
   scrolly = -Self->YPosition;

   itemx = node->XCoord + Self->XPosition;
   itemy = node->YCoord + Self->YPosition;

   if (itemx < 0) scrollx = node->XCoord; /* Scroll left */
   else if ((itemx + node->Width) > Self->AreaWidth) {
      /*** The item is off the right edge of the view ***/
      scrollx = node->XCoord; /* Realign the item to the left of the view */
      if (node->Width < Self->AreaWidth) scrollx -= Self->AreaWidth - node->Width; /* If the item is smaller than the view width, keep the item right-aligned */
   }

   if (itemy < Self->ColumnHeight) {
      scrolly = node->YCoord - Self->ColumnHeight; /* Scroll up */
   }
   else if ((itemy + node->Height) > Self->AreaHeight) {
      scrolly = (node->YCoord + node->Height) - Self->AreaHeight; /* Scroll down */
   }

   if ((scrollx != -Self->XPosition) OR (scrolly != -Self->YPosition)) {
      ActionTags(AC_ScrollToPoint, Self, scrollx, scrolly, 0.0, STF_X|STF_Y);
      CalculateHScroll(Self);
      CalculateVScroll(Self);
   }
}

/******************************************************************************
** Function: DrawItem()
*/

static void DrawItem(struct View *Self, struct XMLTag *Tag)
{
   struct ViewNode *node;
   struct acDraw draw;

   if (Tag) {
      node = Tag->Private;
      draw.XCoord = Self->AreaX + node->XCoord + Self->XPosition;
      draw.YCoord = Self->AreaY + node->YCoord + Self->YPosition;
      draw.Width  = node->Width;
      draw.Height = node->Height;
      ActionMsg(AC_Draw, Self->DrawableID, &draw);
   }
}

/*****************************************************************************
** Internal: SelectItem()
**
** Marks an item as selected and then partially redraws the View in order to
** show the selection.
**
** You can also use this function to deselect all tags, by passing a NULL
** pointer in the Tag argument.
*/

static BYTE SelectItem(struct View *Self, struct XMLTag *Tag, BYTE MultiSelect)
{
   struct XMLTag *scan, **xmllist;
   struct ViewNode *node;
   WORD i;
   BYTE new_selection;

   if (Tag) {
      Self->ActiveTag = Tag->Index;
      Self->SelectedTag = -1;
   }
   else {
       Self->ActiveTag = -1;
       Self->SelectedTag = -1;
   }

   if ((GetField(Self->XML, FID_Tags, FT_POINTER, &xmllist) IS ERR_Okay) AND (xmllist)) {
      /* If we're in single-select mode or we are deselecting everything, scan for any
      ** existing selections and turn them off.
      */

      if ((!Tag) OR (MultiSelect IS FALSE) OR (!(Self->Flags & VWF_MULTISELECT))) {
         for (scan=xmllist[0]; scan; scan=scan->Next) {
            node = scan->Private;
            if (node->Flags & NODE_SELECTED) {
               node->Flags &= ~NODE_SELECTED;
               DrawItem(Self, scan);
            }
         }
      }

      if (Tag) {
         /*** Select the new item ***/

         node = Tag->Private;
         if (!(node->Flags & NODE_SELECTED)) {
            node->Flags |= NODE_SELECTED;
            DrawItem(Self, Tag);
            new_selection = TRUE;
         }
         else new_selection = FALSE;

         SetFields(Self, FID_SelectedTag|TLONG, Tag->Index, TAGEND);

         /*** Ensure that the newly selected item is visible ***/

         CheckItemVisible(Self, xmllist[Self->ActiveTag]);

         /*** Respond to the selection ***/

         if ((new_selection) AND (Self->Flags & VWF_SENSITIVE)) {
            /*** Sensitive mode means that we have to activate whenver a new item is selected ***/
            for (i=0; i < Tag->TotalAttributes; i++) {
               if (StrCompare(Tag->Attributes[i].Name, "insensitive", 0, STR_MATCHLENGTH) IS ERR_Okay)  break;
            }
            if (i >= Tag->TotalAttributes) {
               Action(AC_Activate, Self, NULL);
               return(TRUE);
            }
         }
      }
   }

   return(FALSE);
}

static ERROR SetIcon(struct View *Self, struct XMLTag *Tag, STRING IconFile)
{
   struct ViewNode *node;
   struct RenderInfo info;
   struct Picture *picture;
   UBYTE buffer[512];
   WORD i;

   if (!(node = Tag->Private)) return(ERR_Okay);

   if ((node->Icon) AND (!(node->Flags & NODE_CACHEICON))) acFree(node->Icon);
   node->Icon = NULL;

   if (StrCompare("icons:", IconFile, 6, NULL) IS ERR_Okay) {
      StrFormat(buffer, sizeof(buffer), "%s(16)", IconFile);
   }
   else StrFormat(buffer, sizeof(buffer), "icons:%s(16)", IconFile);

   for (i=0; i < ARRAYSIZE(Self->Cache); i++) {
      if (StrCompare(buffer, Self->Cache[i].String, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         node->Icon = Self->Cache[i].Picture;
         node->Flags |= NODE_CACHEICON;
         return(ERR_Okay);
      }
   }

   if (GetRenderInfo(Self->DrawableID, &info, sizeof(info)) != ERR_Okay) info.BitsPerPixel = 32;

   #ifndef DEBUG
   DebugState(FALSE, -1);
   #endif

   if (NewPrivateObject(ID_PICTURE, NF_CHILD, &picture) IS ERR_Okay) {
      SetFields(picture, FID_Location|TSTRING, buffer, TAGEND);
      picture->Bitmap->BitsPerPixel = info.BitsPerPixel;
      picture->Flags |= PCF_32BITALPHA|PCF_CACHE;
      if (acInit(picture) IS ERR_Okay) {
         node->Icon = picture;
      }
      else acFree(picture);
   }
   else ObjectError(Self, ERH_Refresh, ERR_NewObject);

   #ifndef DEBUG
   DebugState(TRUE, -1);
   #endif

   return(ERR_Okay);
}

